"""
Solar Thermal Climate System (STCS) at Karsruhe University of Applied Sciences.

Adrian Buerger, 2022
Adapted by Wim Van Roy and Andrea Ghezzi, 2023
"""

import numpy as np
import casadi as ca

from camino.settings import GlobalSettings
from camino.utils.cache import CachedFunction
from collections import OrderedDict
import logging

logger = logging.getLogger(__name__)


class System:
    """System dynamics."""

    @staticmethod
    def get_default_initial_state():
        """Get default initial states."""
        return {
            "T_hts": [70.0, 65.0, 63.0, 60.0],
            "T_lts": 14.0,
            "T_fpsc": 20.0,
            "T_fpsc_s": 20.0,
            "T_vtsc": 22.0,
            "T_vtsc_s": 22.0,
            "T_pscf": 18.0,
            "T_pscr": 20.0,
            "T_shx_psc": [11.0, 11.0, 11.0, 11.0],
            "T_shx_ssc": [32.0, 32.0, 32.0, 32.0],
        }

    def _setup_system_dimensions(self):
        self.nx = 19
        self.nb = 3
        self.nu = 6
        self.nc = 6

        self.n_s_ac_lb = 2
        self.n_s_ac_ub = 2

    def _setup_system_components(self):
        # Model parameters
        self.p = {
            # Media
            "rho_w": 1.0e3,
            "c_w": 4.182e3,
            "rho_sl": 1.0e3,
            "c_sl": 3.8e3,
            # Storages
            "V_hts": 2.0,
            "V_lts": 1.0,
            "lambda_hts": [1.15482, 2.89951, 1.195370, 1.000000],
            "eps_hts": 1e-2,
            # Cooling system
            "dT_lc": 4.0,
            "mdot_lc_max": 0.5,
            # Flat plate collectors
            "eta_fpsc": 0.857461,
            "A_fpsc": 53.284,
            "C_fpsc": 248552.0,
            "alpha_fpsc": 1.97512,
            "V_fpsc_s": 2.0e-3,
            "lambda_fpsc_s": 1.0,
            # Tube collectors
            "eta_vtsc": 0.835968,
            "A_vtsc": 31.136,
            "C_vtsc": 445931.0,
            "alpha_vtsc": 1.76747,
            "V_vtsc_s": 2.0e-3,
            "lambda_vtsc_s": 1.0,
            # Pipes connecting solar collectors and solar heat exchanger
            "lambda_psc": 133.144,
            "C_psc": 378160.0,
            # Solar heat exchanger
            "V_shx": 3.8e-3,
            "A_shx": 4.02,
            "alpha_shx": 22535.7,
            "mdot_ssc_max": 0.625,
            # ACM
            "mdot_ac_ht": 0.693,
            "T_ac_ht_min": 55.0,
            "T_ac_ht_max": 95.0,
            "dT_ht_min": 5.0,
            "mdot_ac_lt": 0.8,
            "T_ac_lt_min": 10.0,
            "T_ac_lt_max": 26.0,
            "dT_lt_min": 2.0,
            "T_ac_mt_min": 15.0,
            "T_ac_mt_max": 40.0,
            "Qdot_ac_lb": 0.1,
            "Qdot_ac_ub": 15.0,
            "COP_ac_lb": 0.1,
            "COP_ac_ub": 0.6,
            # HP
            "mdot_hp_lt": 0.5,
            # Recooling tower
            "dT_rc": 2.0,
            # Free cooling
            "mdot_fc_lt": 0.693,
            # PV collectors
            "P_pv_p": 2.0,
        }

        # States

        self.x_index = OrderedDict(
            [
                ("T_hts", [0, 1, 2, 3]),
                ("T_lts", 4),
                ("T_fpsc", 5),
                ("T_fpsc_s", 6),
                ("T_vtsc", 7),
                ("T_vtsc_s", 8),
                ("T_pscf", 9),
                ("T_pscr", 10),
                ("T_shx_psc", [11, 12, 13, 14]),
                ("T_shx_ssc", [15, 16, 17, 18]),
            ]
        )

        self.x_aux_index = {}

        # Continuous controls

        self.u_index = OrderedDict(
            [
                ("v_ppsc", 0),
                ("p_mpsc", 1),
                ("v_pssc", 2),
                ("P_g", 3),
                ("mdot_o_hts_b", 4),
                ("mdot_i_hts_b", 5),
            ]
        )

        # Discrete controls

        self.b_index = OrderedDict(
            [
                ("b_ac", 0),
                ("b_fc", 1),
                ("b_hp", 2),
            ]
        )

        # Time-varying parameters

        self.c_index = OrderedDict(
            [
                ("T_amb", 0),
                ("I_fpsc", 1),
                ("I_vtsc", 2),
                ("Qdot_c", 3),
                ("P_pv_kWp", 4),
                ("p_g", 5),
            ]
        )

    def _setup_operation_specifications_and_limits(self):

        self.p_op = {
            "T": {"min": 8, "max": 98.0},
            "T_lts": {"max": 18.0},
            "T_sc": {
                "T_sc_so": 65.0,
                "M_sc_so": 60.0,
                "v_ppsc_so": 1.0,
                "T_feed_max": 85.0,
            },
            "p_mpsc": {
                "min_mpc": 0.25,
                "min_real": 0.25,
                "max": 0.884,
            },
            "v_ppsc": {"min_mpc": 0.0, "min_real": 0.3, "max": 1.0, "Pmax": 4e2},
            "dp_mpsc": {
                "Pmax": 1e0,
            },
            "dp_mssc": {
                "Pmax": 1e0,
            },
            "dp_macm": {
                "Pmax": 1e0,
            },
            "v_pssc": {"min_mpc": 0.0, "min_real": 0.3, "max": 1.0, "Pmax": 3e2},
            "acm": {
                "min_up_time": [3600.0, 900.0],
                "min_down_time": [1800.0, 900.0],
                "Pmax": [1.5e3, 1.2e3],
                # "Pmax": [1.5e3 + 1.6e3, 1.2e3 + 1.6e3]
            },
            "hp": {"min_up_time": [3600.0], "min_down_time": [1800.0]},
            "grid": {"P_g_max": 1e4},
        }

    def _setup_simulation_control_parameters(self):

        self.p_csim = {
            "dT_ac_ht": 5.0,
            "dT_ac_lt": 1.0,
            "dT_ac_mt": 1.0,
            "dT_sc_ub": 15.0,
            "dT_sc_lb": 5.0,
            "dT_vtsc_fpsc_ub": 5.0,
            "dT_vtsc_fpsc_lb": -5.0,
            "dT_o_hts_b_ub": 10.0,
            "dT_o_hts_b_lb": 0.0,
            "T_i_hts_b_active": 80.0,
            "dT_i_hts_b_ub": 10.0,
            "dT_i_hts_b_lb": 0.0,
        }

    def __init__(self):
        """Create system."""
        logger.debug("Creating system")
        self._setup_system_dimensions()
        self._setup_system_components()
        self._setup_operation_specifications_and_limits()
        self._setup_simulation_control_parameters()
        logger.debug("Creating system variables")
        self._setup_model_variables()
        logger.debug("System initialized")

    def _setup_model_variables(self):
        """Set up variables."""
        self.x = GlobalSettings.CASADI_VAR.sym("x", self.nx)
        self.b = GlobalSettings.CASADI_VAR.sym("b", self.nb)
        self.u = GlobalSettings.CASADI_VAR.sym("u", self.nu)
        self.c = GlobalSettings.CASADI_VAR.sym("c", self.nc)

    def get_f_fcn(self):
        """Create solar thermal climate system model."""
        logger.debug("Creating system model")
        # States
        T_hts = self.x[self.x_index["T_hts"]]
        T_lts = self.x[self.x_index["T_lts"]]

        T_fpsc = self.x[self.x_index["T_fpsc"]]
        T_fpsc_s = self.x[self.x_index["T_fpsc_s"]]
        T_vtsc = self.x[self.x_index["T_vtsc"]]
        T_vtsc_s = self.x[self.x_index["T_vtsc_s"]]

        T_pscf = self.x[self.x_index["T_pscf"]]
        T_pscr = self.x[self.x_index["T_pscr"]]

        T_shx_psc = self.x[self.x_index["T_shx_psc"]]
        T_shx_ssc = self.x[self.x_index["T_shx_ssc"]]

        # Discrete controls
        b_ac = self.b[self.b_index["b_ac"]]
        b_fc = self.b[self.b_index["b_fc"]]
        b_hp = self.b[self.b_index["b_hp"]]

        # Continuous controls
        v_ppsc = self.u[self.u_index["v_ppsc"]]
        p_mpsc = self.u[self.u_index["p_mpsc"]]
        v_pssc = self.u[self.u_index["v_pssc"]]

        mdot_o_hts_b = self.u[self.u_index["mdot_o_hts_b"]]
        mdot_i_hts_b = self.u[self.u_index["mdot_i_hts_b"]]

        # Time-varying parameters
        T_amb = self.c[self.c_index["T_amb"]]
        I_vtsc = self.c[self.c_index["I_vtsc"]]
        I_fpsc = self.c[self.c_index["I_fpsc"]]

        Qdot_c = self.c[self.c_index["Qdot_c"]]

        # Modeling

        f = []

        # Grey box ACM model (spline version)

        def qdot_ac(T_lts, T_hts, T_mts):

            if (
                (T_lts < self.p["T_ac_lt_min"] - self.p["dT_lt_min"])
                or (T_hts < self.p["T_ac_ht_min"] - self.p["dT_ht_min"])
                or (T_mts > self.p["T_ac_mt_max"])
            ):

                return self.p["Qdot_ac_lb"]

            T_mts = max(T_mts, self.p["T_ac_mt_min"])

            char_curve_acm = np.asarray(
                [
                    1.0,
                    T_lts,
                    T_hts,
                    (T_mts),
                    T_lts**2,
                    T_hts**2,
                    (T_mts) ** 2,
                    T_lts * T_hts,
                    T_lts * (T_mts),
                    T_hts * (T_mts),
                    T_lts * T_hts * (T_mts),
                ]
            )

            params_Qdot_ac_lt = np.asarray(
                [
                    -5.6924,
                    1.12102,
                    0.291654,
                    -0.484546,
                    -0.00585722,
                    -0.00140483,
                    0.00795341,
                    0.00399118,
                    -0.0287113,
                    -0.00221606,
                    5.42825e-05,
                ]
            )

            return np.dot(char_curve_acm, params_Qdot_ac_lt)

        def cop_ac(T_lts, T_hts, T_mts):

            if (
                (T_lts < self.p["T_ac_lt_min"] - self.p["dT_lt_min"])
                or (T_hts < self.p["T_ac_ht_min"] - self.p["dT_ht_min"])
                or (T_mts > self.p["T_ac_mt_max"])
            ):

                return self.p["COP_ac_lb"]

            T_mts = max(T_mts, self.p["T_ac_mt_min"])

            char_curve_acm = np.asarray(
                [
                    1.0,
                    T_lts,
                    T_hts,
                    (T_mts),
                    T_lts**2,
                    T_hts**2,
                    (T_mts) ** 2,
                    T_lts * T_hts,
                    T_lts * (T_mts),
                    T_hts * (T_mts),
                    T_lts * T_hts * (T_mts),
                ]
            )

            params_COP_ac = np.asarray(
                [
                    2.03268,
                    -0.116526,
                    -0.0165648,
                    -0.043367,
                    -0.00074309,
                    -0.000105659,
                    -0.00172085,
                    0.00113422,
                    0.00540221,
                    0.00116735,
                    -3.87996e-05,
                ]
            )

            return np.dot(char_curve_acm, params_COP_ac)

        def setup_acm_spline(name, data_fcn, lb, ub):

            T_lts = np.arange(-4, 50, 2)
            T_hts = np.arange(0, 140, 2)
            T_mts = np.arange(-5, 45, 2)

            data = []

            for T_lts_k in T_lts:
                for T_hts_l in T_hts:
                    for T_mts_m in T_mts:

                        T_lts_s = max(
                            self.p["T_ac_lt_min"], min(self.p["T_ac_lt_max"], T_lts_k)
                        )
                        T_hts_s = max(
                            self.p["T_ac_ht_min"], min(self.p["T_ac_ht_max"], T_hts_l)
                        )
                        T_mts_s = max(
                            self.p["T_ac_mt_min"], min(self.p["T_ac_mt_max"], T_mts_m)
                        )

                        if (
                            (T_lts_k < self.p["T_ac_lt_min"])
                            or (T_hts_l < self.p["T_ac_ht_min"])
                            or (T_mts_m > self.p["T_ac_mt_max"])
                        ):
                            data.append(lb)
                        else:
                            data.append(
                                data_fcn(T_lts=T_lts_s, T_hts=T_hts_s, T_mts=T_mts_s)
                            )

            data = np.asarray(data)
            data[data <= lb] = lb
            data[data >= ub] = ub

            return ca.interpolant(name, "bspline", [T_mts, T_hts, T_lts], data)

        iota_qdot = setup_acm_spline(
            "qdot", qdot_ac, self.p["Qdot_ac_lb"], self.p["Qdot_ac_ub"]
        )
        iota_cop = setup_acm_spline(
            "cop", cop_ac, self.p["COP_ac_lb"], self.p["COP_ac_ub"]
        )

        Qdot_ac_lt = (
            iota_qdot(ca.veccat(T_amb + self.p["dT_rc"], T_hts[0], T_lts)) * 1e3
        )
        # Better would be to limit COP_ac!
        COP_ac = iota_cop(ca.veccat(T_amb + self.p["dT_rc"], T_hts[0], T_lts))

        # TODO: Possible error
        T_ac_ht = T_hts[0] - (
            (Qdot_ac_lt)
            / ca.fmax(COP_ac, 1e-6)
            / (self.p["mdot_ac_ht"] * self.p["c_w"])
        )
        T_ac_lt = T_lts - (Qdot_ac_lt / (self.p["mdot_ac_lt"] * self.p["c_w"]))

        # Free cooling

        T_fc_lt = T_amb + self.p["dT_rc"]

        # Heat pump

        Qdot_hp_lt = (
            0.35246631376323156 * T_lts
            - 0.07341224489795989 * (T_amb + self.p["dT_rc"])
            + 10.882680173898295
        ) * 1e3
        T_hp_lt = T_lts - (Qdot_hp_lt / (self.p["mdot_hp_lt"] * self.p["c_w"]))

        # HT storage model

        mdot_ssc = self.p["mdot_ssc_max"] * v_pssc

        # Parameter hts:
        m_hts = (self.p["V_hts"] * self.p["rho_w"]) / T_hts.numel()

        mdot_hts_t_s = mdot_ssc - b_ac * self.p["mdot_ac_ht"]
        mdot_hts_t_sbar = ca.sqrt(mdot_hts_t_s**2 + self.p["eps_hts"])

        mdot_hts_b_s = mdot_i_hts_b - mdot_o_hts_b
        mdot_hts_b_sbar = ca.sqrt(mdot_hts_b_s**2 + self.p["eps_hts"])

        f.append(
            (1.0 / m_hts)
            * (
                mdot_ssc * T_shx_ssc[-1]
                - b_ac * self.p["mdot_ac_ht"] * T_hts[0]
                - (
                    mdot_hts_t_s * ((T_hts[0] + T_hts[1]) / 2.0)
                    + mdot_hts_t_sbar * ((T_hts[0] - T_hts[1]) / 2.0)
                )
                - (self.p["lambda_hts"][0] / self.p["c_w"] * (T_hts[0] - T_amb))
            )
        )

        f.append(
            (1.0 / m_hts)
            * (
                (b_ac * self.p["mdot_ac_ht"] - mdot_i_hts_b) * T_ac_ht
                - (mdot_ssc - mdot_o_hts_b) * T_hts[1]
                + (
                    mdot_hts_t_s * ((T_hts[0] + T_hts[1]) / 2.0)
                    + mdot_hts_t_sbar * ((T_hts[0] - T_hts[1]) / 2.0)
                )
                + (
                    mdot_hts_b_s * ((T_hts[2] + T_hts[1]) / 2.0)
                    + mdot_hts_b_sbar * ((T_hts[2] - T_hts[1]) / 2.0)
                )
                - (self.p["lambda_hts"][3] / self.p["c_w"] * (T_hts[1] - T_amb))
            )
        )

        f.append(
            (1.0 / m_hts)
            * (
                mdot_hts_b_s
                * (((T_hts[-1] + T_hts[-2]) / 2.0) - ((T_hts[-2] + T_hts[-3]) / 2.0))
                + mdot_hts_b_sbar
                * (((T_hts[-1] - T_hts[-2]) / 2.0) - ((T_hts[-2] - T_hts[-3]) / 2.0))
                - (self.p["lambda_hts"][-2] / self.p["c_w"] * (T_hts[-2] - T_amb))
            )
        )

        f.append(
            (1.0 / m_hts)
            * (
                mdot_i_hts_b * T_ac_ht
                - mdot_o_hts_b * T_hts[-1]
                - (
                    mdot_hts_b_s * ((T_hts[-1] + T_hts[-2]) / 2.0)
                    + mdot_hts_b_sbar * ((T_hts[-1] - T_hts[-2]) / 2.0)
                )
                - (self.p["lambda_hts"][-1] / self.p["c_w"] * (T_hts[-1] - T_amb))
            )
        )

        # LT storage and cooling system model

        mdot_lc = Qdot_c / (self.p["c_w"] * self.p["dT_lc"])

        m_lts = self.p["V_lts"] * self.p["rho_w"]

        f.append(
            (1.0 / m_lts)
            * (
                mdot_lc * self.p["dT_lc"]
                + b_ac * self.p["mdot_ac_lt"] * (T_ac_lt - T_lts)
                + b_fc * self.p["mdot_fc_lt"] * (T_fc_lt - T_lts)
                + b_hp * self.p["mdot_hp_lt"] * (T_hp_lt - T_lts)
            )
        )

        # Flat plate collectors

        data_v_ppsc = [-0.1, 0.0, 0.4, 0.6, 0.8, 1.0, 1.1]
        data_p_mpsc = [0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]

        data_mdot_fpsc = np.array(
            [
                0.00116667,
                0.0,
                0.02765,
                0.03511667,
                0.04258333,
                0.04993333,
                0.04993333,
                0.00116667,
                0.0,
                0.02765,
                0.03511667,
                0.04258333,
                0.04993333,
                0.04993333,
                0.0035,
                0.0,
                0.08343,
                0.11165333,
                0.13568,
                0.15895333,
                0.15895333,
                0.005,
                0.0,
                0.12508167,
                0.16568,
                0.20563333,
                0.2397,
                0.2397,
                0.0055,
                0.0,
                0.13999333,
                0.1859,
                0.22790167,
                0.26488,
                0.26488,
                0.006,
                0.0,
                0.14969167,
                0.19844,
                0.24394167,
                0.28695333,
                0.28695333,
                0.00633333,
                0.0,
                0.15706667,
                0.21070833,
                0.25807,
                0.310775,
                0.310775,
                0.0075,
                0.0,
                0.17047833,
                0.229775,
                0.28826667,
                0.34190833,
                0.34190833,
                0.0095,
                0.0,
                0.20687333,
                0.27500667,
                0.331775,
                0.37235333,
                0.37235333,
                0.013,
                0.0,
                0.24111667,
                0.31581,
                0.38300833,
                0.44705167,
                0.44705167,
                0.013,
                0.0,
                0.24111667,
                0.31581,
                0.38300833,
                0.44705167,
                0.44705167,
            ]
        )

        iota_mdot_fpsc = ca.interpolant(
            "iota_mdot_fpsc", "bspline", [data_v_ppsc, data_p_mpsc], data_mdot_fpsc
        )

        mdot_fpsc = iota_mdot_fpsc(ca.veccat(v_ppsc, p_mpsc))

        Qdot_fpsc = self.p["eta_fpsc"] * self.p["A_fpsc"] * I_fpsc
        Qdot_fpsc_amb = self.p["alpha_fpsc"] * self.p["A_fpsc"] * (T_fpsc - T_amb)

        f.append(
            (1.0 / self.p["C_fpsc"])
            * (
                mdot_fpsc * self.p["c_sl"] * (T_pscf - T_fpsc)
                + Qdot_fpsc
                - Qdot_fpsc_amb
            )
        )

        f.append(
            (1.0 / (self.p["V_fpsc_s"] * self.p["rho_sl"] * self.p["c_sl"]))
            * (
                mdot_fpsc * self.p["c_sl"] * (T_fpsc - T_fpsc_s)
                - self.p["lambda_fpsc_s"] * (T_fpsc_s - T_amb)
            )
        )

        # Tube collectors

        data_mdot_vtsc = np.array(
            [
                0.0155,
                0.0,
                0.36735,
                0.46655,
                0.56575,
                0.6634,
                0.6634,
                0.0155,
                0.0,
                0.36735,
                0.46655,
                0.56575,
                0.6634,
                0.6634,
                0.01316667,
                0.0,
                0.32157,
                0.41501333,
                0.50432,
                0.59438,
                0.59438,
                0.01166667,
                0.0,
                0.29325167,
                0.37932,
                0.4577,
                0.54363333,
                0.54363333,
                0.01116667,
                0.0,
                0.28167333,
                0.3641,
                0.44043167,
                0.52345333,
                0.52345333,
                0.01066667,
                0.0,
                0.271975,
                0.34822667,
                0.42439167,
                0.50138,
                0.50138,
                0.01033333,
                0.0,
                0.25626667,
                0.33095833,
                0.39859667,
                0.464225,
                0.464225,
                0.00916667,
                0.0,
                0.217855,
                0.275225,
                0.3384,
                0.39975833,
                0.39975833,
                0.00716667,
                0.0,
                0.15479333,
                0.19832667,
                0.243225,
                0.30098,
                0.30098,
                0.00366667,
                0.0,
                0.06721667,
                0.08752333,
                0.10865833,
                0.13128167,
                0.13128167,
                0.00366667,
                0.0,
                0.06721667,
                0.08752333,
                0.10865833,
                0.13128167,
                0.13128167,
            ]
        )

        iota_mdot_vtsc = ca.interpolant(
            "iota_mdot_vtsc", "bspline", [data_v_ppsc, data_p_mpsc], data_mdot_vtsc
        )

        mdot_vtsc = iota_mdot_vtsc(ca.veccat(v_ppsc, p_mpsc))

        Qdot_vtsc = self.p["eta_vtsc"] * self.p["A_vtsc"] * I_vtsc
        Qdot_vtsc_amb = self.p["alpha_vtsc"] * self.p["A_vtsc"] * (T_vtsc - T_amb)

        f.append(
            (1.0 / self.p["C_vtsc"])
            * (
                mdot_vtsc * self.p["c_sl"] * (T_pscf - T_vtsc)
                + Qdot_vtsc
                - Qdot_vtsc_amb
            )
        )

        f.append(
            (1.0 / (self.p["V_vtsc_s"] * self.p["rho_sl"] * self.p["c_sl"]))
            * (
                mdot_vtsc * self.p["c_sl"] * (T_vtsc - T_vtsc_s)
                - self.p["lambda_vtsc_s"] * (T_vtsc_s - T_amb)
            )
        )

        # Pipes connecting solar collectors and solar heat exchanger

        f.append(
            1.0
            / self.p["C_psc"]
            * (
                (mdot_fpsc + mdot_vtsc) * self.p["c_sl"] * (T_shx_psc[-1] - T_pscf)
                - self.p["lambda_psc"] * (T_pscf - T_amb)
            )
        )

        f.append(
            1.0
            / self.p["C_psc"]
            * (
                mdot_fpsc * self.p["c_sl"] * T_fpsc_s
                + mdot_vtsc * self.p["c_sl"] * T_vtsc_s
                - (mdot_fpsc + mdot_vtsc) * self.p["c_sl"] * T_pscr
                - self.p["lambda_psc"] * (T_pscr - T_amb)
            )
        )

        # Solar heat exchanger

        m_shx_psc = self.p["V_shx"] * self.p["rho_sl"] / T_shx_psc.numel()
        m_shx_ssc = self.p["V_shx"] * self.p["rho_w"] / T_shx_psc.numel()

        A_shx_k = self.p["A_shx"] / T_shx_psc.numel()

        f.append(
            (1.0 / (m_shx_psc * self.p["c_sl"]))
            * (
                (mdot_fpsc + mdot_vtsc) * self.p["c_sl"] * (T_pscr - T_shx_psc[0])
                - (A_shx_k * self.p["alpha_shx"] * (T_shx_psc[0] - T_shx_ssc[-1]))
            )
        )

        for k in range(1, T_shx_psc.numel()):

            f.append(
                (1.0 / (m_shx_psc * self.p["c_sl"]))
                * (
                    (mdot_fpsc + mdot_vtsc)
                    * self.p["c_sl"]
                    * (T_shx_psc[k - 1] - T_shx_psc[k])
                    - (
                        A_shx_k
                        * self.p["alpha_shx"]
                        * (T_shx_psc[k] - T_shx_ssc[-1 - k])
                    )
                )
            )

        f.append(
            (1.0 / (m_shx_ssc * self.p["c_w"]))
            * (
                (mdot_ssc - mdot_o_hts_b) * self.p["c_w"] * (T_hts[1] - T_shx_ssc[0])
                + mdot_o_hts_b * self.p["c_w"] * (T_hts[-1] - T_shx_ssc[0])
                + (A_shx_k * self.p["alpha_shx"] * (T_shx_psc[-1] - T_shx_ssc[0]))
            )
        )

        for k in range(1, T_shx_ssc.numel()):

            f.append(
                (1.0 / (m_shx_ssc * self.p["c_w"]))
                * (
                    mdot_ssc * self.p["c_w"] * (T_shx_ssc[k - 1] - T_shx_ssc[k])
                    + (
                        A_shx_k
                        * self.p["alpha_shx"]
                        * (T_shx_psc[-1 - k] - T_shx_ssc[k])
                    )
                )
            )

        return ca.Function("f", [self.x, self.c, self.u, self.b], [ca.veccat(*f)])

    def get_integrator(self):
        """Set up simulator."""
        dt = ca.MX.sym("dt")
        f = CachedFunction("stcs_f", self.get_f_fcn)

        ode = {
            "x": self.x,
            "p": ca.veccat(dt, self.c, self.u, self.b),
            "ode": dt * f(self.x, self.c, self.u, self.b),
        }
        return ca.integrator("integrator", "cvodes", ode, 0.0, 1.0)

    def get_t_ac_min_function(self, use_big_m_constraints=True):
        """
        Create a function to represent the minimum uptime.

        Function has as argument x, c, b, slack_variables
        The output is the T_ac_min constraint (>0).
        """
        # ACM operation condition equations
        s_ac_lb = ca.MX.sym("s_ac_lb", self.n_s_ac_lb)
        if use_big_m_constraints:
            T_ac_min = ca.veccat(
                self.x[self.x_index["T_lts"]]
                - self.p_op["T"]["min"]
                - self.b[self.b_index["b_ac"]]
                * (self.p["T_ac_lt_min"] - self.p_op["T"]["min"])
                + s_ac_lb[0],
                self.x[self.x_index["T_hts"][0]]
                - self.p_op["T"]["min"]
                - self.b[self.b_index["b_ac"]]
                * (self.p["T_ac_ht_min"] - self.p_op["T"]["min"])
                + s_ac_lb[1],
            )
        else:
            T_ac_min = self.b[self.b_index["b_ac"]] * ca.veccat(
                self.x[self.x_index["T_lts"]] - self.p["T_ac_lt_min"] + s_ac_lb[0],
                self.x[self.x_index["T_hts"][0]] - self.p["T_ac_ht_min"] + s_ac_lb[1],
            )

        return ca.Function(
            "T_ac_min_fcn", [self.x, self.c, self.b, s_ac_lb], [T_ac_min]
        )

    def get_t_ac_max_function(self, use_big_m_constraints=True):
        """
        Create a function to represent the maximum uptime.

        Function has as argument x, c, b, slack_variables
        The output is the T_ac_min constraint (>0).
        """
        s_ac_ub = ca.MX.sym("s_ac_ub", self.n_s_ac_ub)
        if use_big_m_constraints:
            T_ac_max = ca.veccat(
                self.x[self.x_index["T_lts"]]
                - self.p_op["T"]["max"]
                - self.b[self.b_index["b_ac"]]
                * (self.p["T_ac_lt_max"] - self.p_op["T"]["max"])
                - s_ac_ub[0],
                self.x[self.x_index["T_hts"][0]]
                - self.p_op["T"]["max"]
                - self.b[self.b_index["b_ac"]]
                * (self.p["T_ac_ht_min"] - self.p_op["T"]["max"])
                - s_ac_ub[1],
            )
        else:
            T_ac_max = self.b[self.b_index["b_ac"]] * ca.veccat(
                self.x[self.x_index["T_lts"]] - self.p["T_ac_lt_max"] - s_ac_ub[0],
                self.x[self.x_index["T_hts"][0]] - self.p["T_ac_ht_max"] - s_ac_ub[1],
            )

        return ca.Function(
            "T_ac_max_fcn", [self.x, self.c, self.b, s_ac_ub], [T_ac_max]
        )

    def get_slacked_state_fcn(self):
        s_x = ca.MX.sym("s_x", self.nx)
        return ca.Function("s_x_fcn", [self.x, s_x], [self.x + s_x])

    def get_v_ppsc_so_fpsc_fcn(self):
        """Get v_ppsc so function."""
        # Assure ppsc is running at high speed when collector temperature is high
        s_ppsc = ca.MX.sym("s_ppsc")
        return ca.Function(
            "v_ppsc_so_fpsc_fcn",
            [self.x, s_ppsc],
            [
                (self.x[self.x_index["T_fpsc"]] - self.p_op["T_sc"]["T_sc_so"])
                * (self.p_op["T_sc"]["v_ppsc_so"] - s_ppsc)
            ],
        )

    def get_v_ppsc_so_vtsc_fcn(self):
        """V ppsc."""
        s_ppsc = ca.MX.sym("s_ppsc")
        return ca.Function(
            "v_ppsc_so_vtsc_fcn",
            [self.x, s_ppsc],
            [
                (self.x[self.x_index["T_vtsc"]] - self.p_op["T_sc"]["T_sc_so"])
                * (self.p_op["T_sc"]["v_ppsc_so"] - s_ppsc)
            ],
        )

    def get_v_ppsc_so_fcn(self):
        """Get v ppsc so function."""
        s_ppsc = ca.MX.sym("s_ppsc")
        return ca.Function(
            "v_ppsc_so_fcn", [self.u, s_ppsc], [s_ppsc - self.u[self.u_index["v_ppsc"]]]
        )

    def get_mdot_hts_b_max_fcn(self):
        """Get mdot hts b max function."""
        # Assure HTS bottom layer mass flows are always smaller or equal to
        # the corresponding total pump flow

        mdot_hts_b_max = ca.veccat(
            self.u[self.u_index["mdot_o_hts_b"]]
            - self.p["mdot_ssc_max"] * self.u[self.u_index["v_pssc"]],
            self.u[self.u_index["mdot_i_hts_b"]]
            - self.b[self.b_index["b_ac"]] * self.p["mdot_ac_ht"],
        )

        return ca.Function("mdot_hts_b_max_fcn", [self.u, self.b], [mdot_hts_b_max])

    def get_electric_power_balance_fcn(self):
        """Get electric power balance fcn."""
        P_hp = (
            0.006381135707410529 * self.x[self.x_index["T_lts"]]
            + 0.06791020408163258 * (self.c[self.c_index["T_amb"]] + self.p["dT_rc"])
            + 1.7165550470428876
        ) * 1e3  # + 8e2

        electricity_consumption = (
            self.p_op["acm"]["Pmax"][0] * self.b[self.b_index["b_ac"]]
            + self.p_op["acm"]["Pmax"][1] * self.b[self.b_index["b_fc"]]
            + P_hp * self.b[self.b_index["b_hp"]]
            + self.p_op["v_ppsc"]["Pmax"] * self.u[self.u_index["v_ppsc"]]
            + self.p_op["v_pssc"]["Pmax"] * self.u[self.u_index["v_pssc"]]
        )

        electric_power_balance = (
            -electricity_consumption
            + self.p["P_pv_p"] * self.c[self.c_index["P_pv_kWp"]]
            + self.p_op["grid"]["P_g_max"] * self.u[self.u_index["P_g"]]
        )

        return ca.Function(
            "electric_power_balance_fcn",
            [self.x, self.u, self.b, self.c],
            [electric_power_balance],
        )

    def get_F1_fcn(self):
        """Get F1."""
        s_ac_lb = ca.MX.sym("s_ac_lb", self.n_s_ac_lb)
        s_ac_ub = ca.MX.sym("s_ac_ub", self.n_s_ac_ub)
        s_x = ca.MX.sym("s_x", self.nx)
        u_prev = ca.MX.sym("u_prev", self.nu)

        F1 = ca.veccat(
            5e0 * s_ac_lb,
            5e0 * s_ac_ub,
            1e1 * s_x,
            1e2 * s_x[self.x_index["T_lts"]],
            self.p_op["dp_mpsc"]["Pmax"]
            * (u_prev[self.u_index["p_mpsc"]] - self.u[self.u_index["p_mpsc"]]),
            self.p_op["dp_mssc"]["Pmax"]
            * (
                u_prev[self.u_index["mdot_o_hts_b"]]
                - self.u[self.u_index["mdot_o_hts_b"]]
            ),
            self.p_op["dp_macm"]["Pmax"]
            * (
                u_prev[self.u_index["mdot_i_hts_b"]]
                - self.u[self.u_index["mdot_i_hts_b"]]
            ),
        )

        return ca.Function("F1_fcn", [s_ac_lb, s_ac_ub, s_x, self.u, u_prev], [F1])

    def get_F2_fcn(self):
        """Get F2."""
        F2 = (
            self.c[self.c_index["p_g"]]
            * self.p_op["grid"]["P_g_max"]
            * self.u[self.u_index["P_g"]]
        )

        return ca.Function("F2_fcn", [self.u, self.c], [F2])

    def get_system_dynamics_collocation(system, d=2):
        """Create collocation."""
        tau_root = [0] + ca.collocation_points(d, "radau")

        f = CachedFunction("stcs_f", system.get_f_fcn)
        C = np.zeros((d + 1, d + 1))
        D = np.zeros(d + 1)

        for j in range(d + 1):
            p = np.poly1d([1])
            for r in range(d + 1):
                if r != j:
                    p *= np.poly1d([1, -tau_root[r]]) / (tau_root[j] - tau_root[r])

            D[j] = p(1.0)
            pder = np.polyder(p)
            for r in range(d + 1):
                C[j, r] = pder(tau_root[r])

        # Collocation equations
        x_k_c = [ca.MX.sym("x_k_c_" + str(j), system.nx) for j in range(d + 1)]
        x_k_next_c = ca.MX.sym("x_k_next_c", system.nx)
        c_k_c = ca.MX.sym("c_k_c", system.nc)
        u_k_c = ca.MX.sym("u_k_c", system.nu)
        b_k_c = ca.MX.sym("b_k_c", system.nb)
        dt_k_c = ca.MX.sym("dt_k_c")

        eq_c = []
        for j in range(1, d + 1):
            x_p_c = 0
            for r in range(d + 1):
                x_p_c += C[r, j] * x_k_c[r]
            f_k_c = f(x_k_c[j], c_k_c, u_k_c, b_k_c)
            eq_c.append(dt_k_c * f_k_c - x_p_c)

        eq_c = ca.veccat(*eq_c)
        xf_c = 0

        for r in range(d + 1):
            xf_c += D[r] * x_k_c[r]
            eq_d = xf_c - x_k_next_c

        return ca.Function(
            "F",
            x_k_c + [x_k_next_c, c_k_c, u_k_c, b_k_c, dt_k_c],
            [eq_c, eq_d],
            ["x_k_" + str(j) for j in range(d + 1)]
            + ["x_k_next", "c_k", "u_k", "b_k", "dt_k"],
            ["eq_c", "eq_d"],
        )


if __name__ == "__main__":
    system = System()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Global settings."""

from typing import Any, Dict
from dataclasses import dataclass, field
from os import path, makedirs, environ
import casadi as ca


def create_and_return(folder):
    if not path.exists(folder):
        makedirs(folder)
    return folder


@dataclass
class _GlobalSettings:
    _lock: bool = False
    _CASADI_VAR: Any = ca.MX
    SOURCE_FOLDER: str = path.dirname(path.abspath(__file__))
    _DATA_FOLDER: str = path.join(SOURCE_FOLDER, "../data")
    _IMG_DIR: str = path.join(SOURCE_FOLDER, "../results/figures")
    _OUT_DIR: str = path.join(SOURCE_FOLDER, "../results")
    _CACHE_FOLDER: str = path.join(SOURCE_FOLDER, "../data/cache")

    @property
    def CASADI_VAR(self):
        self._lock = True
        return self._CASADI_VAR

    @CASADI_VAR.setter
    def CASADI_VAR(self, value: Any):
        if self._lock:
            raise RuntimeError(
                "Casadi var already used! You can not change it anymore!"
            )
        self._CASADI_VAR = value

    @property
    def IMG_DIR(self):
        return create_and_return(self._IMG_DIR)

    @IMG_DIR.setter
    def IMG_DIR(self, value):
        """Set image directory."""
        self._IMG_DIR = value

    @property
    def OUT_DIR(self):
        return create_and_return(self._OUT_DIR)

    @OUT_DIR.setter
    def OUT_DIR(self, value):
        """Set output dir directory."""
        self._OUT_DIR = value

    @property
    def CACHE_FOLDER(self):
        return create_and_return(self._CACHE_FOLDER)

    @CACHE_FOLDER.setter
    def CACHE_FOLDER(self, value):
        """Set output dir directory."""
        self._CACHE_FOLDER = value

    @property
    def DATA_FOLDER(self):
        return create_and_return(self._DATA_FOLDER)

    @DATA_FOLDER.setter
    def DATA_FOLDER(self, value):
        """Set output dir directory."""
        self._DATA_FOLDER = value


GlobalSettings = _GlobalSettings()


@dataclass(init=True)
class Settings:
    from camino.utils.conversion import to_bool

    TIME_LIMIT: float = 3600.0
    TIME_LIMIT_SOLVER_ONLY: bool = False
    WITH_JIT: bool = False
    WITH_PLOT: bool = False
    INF: float = 1e8
    BIGM: float = 1e6
    EPS: float = 1e-6
    OBJECTIVE_TOL: float = 1e-5
    CONSTRAINT_INT_TOL: float = 1e-2  # Due to rounding, this will be almost EPS
    CONSTRAINT_TOL: float = 1e-4
    BENDERS_LB: float = -1e16
    _MIP_SOLVER: str = "highs"

    WITH_DEBUG: bool = to_bool(environ.get("DEBUG", False))
    WITH_LOG_DATA: bool = to_bool(environ.get("LOG_DATA", False))
    WITH_SAVE_FIG = to_bool(environ.get("SAVE_FIG", False))
    MINLP_TOLERANCE: float = 0.01
    MINLP_TOLERANCE_ABS: float = 0.01
    BRMIQP_GAP: float = 1e-4
    LBMILP_GAP: float = 1e-4
    RHO_AMPLIFICATION: float = 1.5
    ALPHA_KRONQVIST: float = 0.5
    USE_RELAXED_SOL_AS_LINEARIZATION = True
    USE_TIGHT_MIPGAP_FIRST_ITERATION = True
    # WITH_DEFAULT_SETTINGS = to_bool(environ.get("DEFAULT", True))

    PUMP_MAX_STEP_IMPROVEMENTS = 5
    PUMP_MAX_ITER = 1000
    PUMP_MAX_TRY = 10
    PARALLEL_SOLUTIONS = 5

    AMPL_EXPORT_SETTINGS: Dict[str, Any] = field(default_factory=lambda: {})
    IPOPT_SETTINGS: Dict[str, Any] = field(
        default_factory=lambda: {
            "ipopt.linear_solver": "mumps",
        }
    )
    BONMIN_SETTINGS: Dict[str, Any] = field(
        default_factory=lambda: {
            "bonmin.time_limit": Settings.TIME_LIMIT,
            "bonmin.linear_solver": "mumps",
            "bonmin.allowable_gap": Settings.MINLP_TOLERANCE_ABS,
            "bonmin.allowable_fraction_gap": Settings.MINLP_TOLERANCE,
        }
    )
    MIP_SETTINGS_ALL: Dict[str, Dict[str, Any]] = field(
        default_factory=lambda: {
            "highs": {},
            "cbc": {},
            "gurobi": {
                "gurobi.NumericFocus": 1,
                # Note since this is only the dual, and since rounding occurs,
                # this low tolerance does not affect the solution!
                "gurobi.FeasibilityTol": Settings.CONSTRAINT_INT_TOL,
                "gurobi.IntFeasTol": Settings.CONSTRAINT_INT_TOL,
                "gurobi.PoolSearchMode": 1,
                "gurobi.PoolSolutions": 1,
                "gurobi.TimeLimit": 3600,
            },
        }
    )

    def __post_init__(self, *args, **kwargs):
        """Settings."""
        super(Settings, self).__init__(*args, **kwargs)
        if self.WITH_DEBUG:
            self.IPOPT_SETTINGS.update({"ipopt.print_level": 5})
            self.MIP_SETTINGS_ALL["gurobi"].update({"gurobi.output_flag": 1})
        else:
            self.IPOPT_SETTINGS.update({"ipopt.print_level": 0})
            self.MIP_SETTINGS_ALL["gurobi"].update({"gurobi.output_flag": 0})
        self.MIP_SOLVER = environ.get("MIP_SOLVER", "gurobi")

    @property
    def USE_SOLUTION_POOL(self):
        return self.MIP_SETTINGS_ALL["gurobi"]["gurobi.PoolSolutions"] > 1

    @USE_SOLUTION_POOL.setter
    def USE_SOLUTION_POOL(self, value):
        if value:
            self.MIP_SETTINGS_ALL["gurobi"]["gurobi.PoolSearchMode"] = 1
            self.MIP_SETTINGS_ALL["gurobi"]["gurobi.PoolSolutions"] = 5
        else:
            self.MIP_SETTINGS_ALL["gurobi"]["gurobi.PoolSearchMode"] = 0
            self.MIP_SETTINGS_ALL["gurobi"]["gurobi.PoolSolutions"] = 1

    @property
    def MIP_SOLVER(self):
        return self._MIP_SOLVER

    @MIP_SOLVER.setter
    def MIP_SOLVER(self, value):
        if value not in self.MIP_SETTINGS_ALL:
            raise Exception(
                "Configure a MIP_SOLVER from the list: %s"
                % str(", ".join(self.MIP_SETTINGS_ALL.keys()))
            )
        else:
            self._MIP_SOLVER = value

    @property
    def MIP_SETTINGS(self):
        return self.MIP_SETTINGS_ALL[self.MIP_SOLVER]



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Mixed-integer linearity in nonlinear optimization: a trust region approach

Alberto De Marchi 2023

Notes:
    - It seems that taking an L1-norm trust region doesn't work when handling
      problems with integer states. The algorithm is stuck in this case
    - Algorithm is not designed for nonlinear constraints yet. This might give
      problems.
    - The objective function should be linear in the integer variables
"""

from copy import deepcopy
import casadi as ca
import numpy as np
from typing import Tuple
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import GlobalSettings, Settings
from camino.solvers import get_idx_inverse
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.utils import colored
from camino.utils import toc, logging

logger = logging.getLogger(__name__)


class TrustRegionMILP(SolverClass):
    """Create trust region MILP master problem."""

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create benders master MILP."""
        super(TrustRegionMILP, self).__init__(problem, stats, s)
        self.settings = s
        self.options = regularize_options(s.MIP_SETTINGS, {}, s)

        self.f = ca.Function(
            "f", [problem.x, problem.p], [problem.f], {"jit": s.WITH_JIT}
        )
        self.g = ca.Function(
            "g", [problem.x, problem.p], [problem.g], {"jit": s.WITH_JIT}
        )

        self.jac_g = ca.Function(
            "jac_g",
            [problem.x, problem.p],
            [ca.jacobian(problem.g, problem.x)],
            {"jit": s.WITH_JIT},
        )
        self.jac_f = ca.Function(
            "jac_f",
            [problem.x, problem.p],
            [ca.jacobian(problem.f, problem.x)],
            {"jit": s.WITH_JIT},
        )

        self.nr_x = problem.x.numel()
        self.idx_x_c = get_idx_inverse(problem.idx_x_integer, self.nr_x)
        self.nr_x_c = len(self.idx_x_c)
        self.x = GlobalSettings.CASADI_VAR.sym("x", self.nr_x)
        self.options["discrete"] = [
            1 if i in problem.idx_x_integer else 0 for i in range(self.nr_x)
        ]

    def solve(self, nlpdata: MinlpData, delta=1) -> MinlpData:
        """Solve MILP with TR."""
        p = nlpdata.p
        x_hat = nlpdata.x_sol
        dx = self.x - nlpdata.x_sol
        f_lin = self.jac_f(x_hat, p) @ dx
        g_lin = self.g(x_hat, p) + self.jac_g(x_hat, p) @ dx
        # 1 norm - only on the continuous variables...
        g_extra = dx[self.idx_x_c]
        g_extra_lb = -delta * np.ones((self.nr_x_c,))
        g_extra_ub = delta * np.ones((self.nr_x_c,))

        solver = ca.qpsol(
            "milp_tr",
            self.settings.MIP_SOLVER,
            {
                "f": f_lin,
                "g": ca.vertcat(g_lin, g_extra),
                "x": self.x,
            },
            self.options,
        )
        solution = solver(
            x0=x_hat,
            ubx=nlpdata.ubx,
            lbx=nlpdata.lbx,
            ubg=ca.vertcat(nlpdata.ubg, g_extra_ub),
            lbg=ca.vertcat(nlpdata.lbg, g_extra_lb),
        )
        success, _ = self.collect_stats("TR-MILP", solver, solution)
        nlpdata.prev_solution = solution
        nlpdata.solved = success
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""A set of conversion tools."""

import numpy as np


def to_bool(val):
    """String or bool value to bool."""
    if isinstance(val, bool):
        return val
    else:
        return not (val.lower() in ["0", "false", "false", "no"])


def to_float(val):
    """To single float."""
    if isinstance(val, np.ndarray):
        return to_float(val[0])
    elif isinstance(val, list):
        return to_float(val[0])
    return val


def to_0d(array):
    """To zero dimensions."""
    if isinstance(array, np.ndarray):
        ret = array.squeeze()
    elif isinstance(array, list):
        ret = np.array(array).squeeze()
    else:
        ret = array.full().squeeze()
    if ret.size == 1:
        ret = ret.reshape((-1, 1))
    return ret


def convert_to_flat_list(nr, indices, data):
    """Convert data to a flat list."""
    out = np.zeros((nr,))
    for key, indices in indices.items():
        values = data[key]
        if isinstance(indices, list):
            for idx, val in zip(indices, values):
                out[idx] = val
        else:
            out[indices] = values
    return out



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Base algorithm for random feasibility pumps."""

from copy import deepcopy
import numpy as np
import casadi as ca
from camino.utils import colored
from camino.settings import Settings
from camino.solvers.subsolvers.nlp import NlpSolver
from camino.stats import Stats
from camino.data import MinlpData
from camino.problem import MinlpProblem
from camino.solvers.pumps.utils import integer_error, create_rounded_data
from camino.utils import logging
from camino.solvers import MiSolverClass

logger = logging.getLogger(__name__)


class PumpBaseRandom(MiSolverClass):
    """Random pump base algorithm."""

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        pump,
        nlp=None,
    ):
        """Create a solver class."""
        super(PumpBaseRandom, self).__init__(problem, data, stats, settings)
        self.pump = pump
        self.idx_x_integer = problem.idx_x_integer
        if nlp is None:
            nlp = NlpSolver(problem, stats, settings)
        self.nlp = nlp

    def solve(self, nlpdata: MinlpData, integers_relaxed: bool = False) -> MinlpData:
        """Solve the problem."""
        if self.stats.relaxed_solution is None:
            if not integers_relaxed:
                nlpdata = self.nlp.solve(nlpdata)
            self.stats.relaxed_solution = nlpdata
        else:
            nlpdata = self.stats.relaxed_solution

        logger.info("Solver initialized.")
        last_restart = 0
        self.stats["best_iter"] = -1
        done = False
        best_obj = ca.inf
        lb = nlpdata.obj_val
        prev_int_error = ca.inf

        while not done:
            logger.info(f"Starting iteration: {self.stats['iter_nr']}")
            nlpdata = self.pump.solve(nlpdata)
            random_obj_f = float(self.pump.f(nlpdata.x_sol, nlpdata.p))
            lb = min(random_obj_f, lb)

            colored(f"Current random NLP objective: {random_obj_f:.3e}", "blue")
            if random_obj_f < best_obj:
                datarounded = self.nlp.solve(
                    create_rounded_data(nlpdata, self.idx_x_integer), integers_relaxed=False
                )
                if datarounded.solved:
                    logger.debug(
                        f"NLP f={datarounded.obj_val:.3e} (iter {self.stats['iter_nr']}) "
                        f"vs old f={best_obj:.3e} (itr {self.stats['best_iter']})"
                    )
                    self.update_best_solutions(datarounded)
                else:
                    colored("Infeasible")
            else:
                colored("Not better than best found yet")

            int_error = integer_error(nlpdata.x_sol[self.idx_x_integer])

            self.stats["ub"] = best_obj
            self.stats["iter_nr"] += 1
            done = int_error < self.settings.CONSTRAINT_INT_TOL or (
                not np.isinf(best_obj)
                and (
                    (
                        self.stats["iter_nr"] > self.settings.PUMP_MAX_STEP_IMPROVEMENTS
                        and best_obj < random_obj_f
                    )
                    or self.stats["iter_nr"]
                    > self.settings.PUMP_MAX_STEP_IMPROVEMENTS + self.stats["best_iter"]
                )
            )
            retry = (
                self.stats["iter_nr"] - last_restart > self.settings.PUMP_MAX_TRY
                and prev_int_error < int_error
            )
            prev_int_error = int_error
            if not nlpdata.solved or retry:
                if len(self.best_solutions) > 0:
                    done = True
                else:
                    last_restart = self.stats["iter_nr"]
                    self.pump.alpha = 1.0
                    # If progress is frozen (unsolvable), try to fix it!
                    nlpdata = self.pump.solve(self.stats.relaxed)
                    logger.info(
                        f"Current random NLP (restoration): f={random_obj_f:.3e}"
                    )
            if self.stats["iter_nr"] > self.settings.PUMP_MAX_ITER:
                if len(self.best_solutions) > 0:
                    done = True
                else:
                    return self.get_best_solutions(nlpdata)

        return self.get_best_solutions(nlpdata)

    def reset(self, nlpdata: MinlpData):
        """Reset problem data."""

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart."""
        if not nlpdata.relaxed:
            self.update_best_solutions(nlpdata)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""An random direction search NLP solver."""

import casadi as ca
import numpy as np
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import GlobalSettings, Settings
from camino.utils import logging
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class RandomDirectionProjection(SolverClass):
    """
    Create Random Projection Solver.

    This solver solves an NLP problem. This is either relaxed or the binaries are fixed. The NLP aims to satisfy
    original constraint in g and minimize a linear combination of the original objective and the he distance of
    relaxed x_bin from their rounded value in L1-norm. The rounding procedure to compute x_bin is deterministic
    (to closest integer) and the weight vector in the cost function is random.
    """

    def __init__(
        self,
        problem: MinlpProblem,
        stats: Stats,
        s: Settings,
        norm=2,
        penalty_scaling=0.5,
    ):
        super(RandomDirectionProjection, self).__init__(problem, stats, s)
        options = regularize_options(
            s.IPOPT_SETTINGS, {"jit": s.WITH_JIT, "ipopt.max_iter": 5000}, s
        )
        self.penalty_scaling = penalty_scaling

        self.idx_x_integer = problem.idx_x_integer
        x_bin_var = problem.x[self.idx_x_integer]
        self.nr_x_bin = x_bin_var.shape[0]
        penalty = GlobalSettings.CASADI_VAR.sym("penalty", self.nr_x_bin)
        rounded_value = GlobalSettings.CASADI_VAR.sym(
            "rounded_value", x_bin_var.shape[0]
        )
        alpha = GlobalSettings.CASADI_VAR.sym("alpha", 1)
        self.alpha = 1.0
        self.alpha_reduction = 0.9
        obj_val = GlobalSettings.CASADI_VAR.sym("obj_val", 1)
        int_error = GlobalSettings.CASADI_VAR.sym("int_error", 1)

        self.max_rounding_error = x_bin_var.shape[0]
        if norm == 1:
            penalty_term = ca.sum1(ca.fabs(x_bin_var - rounded_value) * penalty)
            self.int_error_calc = lambda x, y: float(ca.norm_1((x - y)))
        else:
            penalty_term = ca.norm_2((x_bin_var - rounded_value) * penalty) ** 2
            self.int_error_calc = lambda x, y: float(ca.norm_2((x - y)) ** 2)

        self.solver = ca.nlpsol(
            "nlpsol",
            "ipopt",
            {
                "f": alpha / obj_val * problem.f
                + (1 - alpha) / int_error * penalty_term,
                "g": problem.g,
                "x": problem.x,
                "p": ca.vertcat(
                    problem.p, rounded_value, penalty, alpha, obj_val, int_error
                ),
            },
            options,
        )
        self.f = ca.Function("f", [problem.x, problem.p], [problem.f])

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        success_out = []
        sols_out = []
        for sol in nlpdata.solutions_all:
            lbx = nlpdata.lbx
            ubx = nlpdata.ubx

            x_bin_var = to_0d(sol["x"][self.idx_x_integer])
            x_rounded = np.round(x_bin_var).flatten()
            int_error = self.int_error_calc(x_bin_var, x_rounded)

            penalty = np.random.rand(len(self.idx_x_integer))
            total = float(ca.sum1(penalty)) / self.nr_x_bin
            logger.info(
                f"{nlpdata.obj_val=:.3} rounding error={int_error:.3f} "
                f"- weight {float(self.alpha):.3e}"
            )
            if self.alpha < 1e-4:
                self.alpha = 0

            new_sol = self.solver(
                x0=nlpdata.x0,
                lbx=lbx,
                ubx=ubx,
                lbg=nlpdata.lbg,
                ubg=nlpdata.ubg,
                p=ca.vertcat(
                    nlpdata.p,
                    x_rounded,
                    penalty,
                    np.array([self.alpha / total, int_error, nlpdata.obj_val]),
                ),
            )

            success, _ = self.collect_stats("RNLP", sol=new_sol)
            if not success:
                logger.debug("Infeasible solution!")
            success_out.append(success)
            sols_out.append(new_sol)

        self.alpha *= self.alpha_reduction
        nlpdata.prev_solutions = sols_out
        nlpdata.solved_all = success_out
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""An easier way to define a problem, similar to casadi opti."""


import copy
from typing import Optional, Union, List, Tuple
from camino.settings import GlobalSettings
from casadi import (
    inf,
    vertcat,
    hessian,
    Function,
    DM,
    reshape,
    jacobian,
    vcat,
    Sparsity,
    blockcat,
)
from math import isinf
from camino.problems import MinlpProblem, MinlpData
import numpy as np


def extract(sol, lst):
    """Extract a value."""
    return [float(sol["x"][i]) for i in lst]


def make_list(value, nr=1):
    """Make list."""
    if isinstance(value, np.ndarray):
        return list(np.reshape(value, (nr,)))
    elif not isinstance(value, list):
        return [value] * nr
    elif isinstance(value, list):
        if len(value) != nr:
            raise Exception("Value error!")
        return value


def count_shape(shape):
    """Shape to nr."""
    if isinstance(shape, (tuple, list)):
        return shape[0] * shape[1]
    else:
        return shape


def as_shape(x, shape):
    """Make shape."""
    if isinstance(shape, (tuple, list)):
        return reshape(x, shape)
    else:
        return x


def make_soc_matrix(h, new_h):
    if isinstance(h, list):
        if h == []:
            return new_h
    else:
        upper_left = copy.deepcopy(h)
        lower_right = copy.deepcopy(new_h)
        upper_right = Sparsity(upper_left.shape[0], lower_right.shape[1])
        lower_left = Sparsity(lower_right.shape[0], upper_left.shape[1])
        return blockcat(upper_left, upper_right, lower_left, lower_right)


class Description:
    """Description for Casadi."""

    def __init__(self):
        """Create description."""
        self.h = []  # for soc-type constraints
        self.g = []
        self.ubg = []
        self.lbg = []
        self.g_lin = []
        self.g_dis = []
        self.g_dwelltime = []  # list of constraints involving dwell time
        self.w = []
        self.w0 = []
        self.indices = {}  # Names with indices
        self.p = []
        self.p0 = []
        self.indices_p = {}  # Names with indices
        self.lbw = []
        self.ubw = []
        self.discrete = []
        self.f = GlobalSettings.CASADI_VAR(0)
        self.r = []  # residual for GN hessian computation
        self.check_illconditioned = False

    def add_w(
        self,
        lb: Union[float, List[float]],
        w: GlobalSettings.CASADI_VAR,
        ub: Union[float, List[float]],
        w0: Optional[Union[float, List[float]]],
        discrete: Union[bool, List[int]] = False,
    ):
        """Add an existing symbolic variable."""
        if self.check_illconditioned:
            too_small = [i for i in lb if abs(i) > 1e5 and not isinf(i)]
            too_large = [i for i in ub if abs(i) > 1e5 and not isinf(i)]

            if len(too_small) > 0 or len(too_large) > 0:
                raise Exception(
                    "Illconditioned {lb} <= {w} <= {ub}".format(ub=ub, lb=lb, w=w)
                )

        idx = [i + len(self.lbw) for i in range(len(lb))]
        self.w += [w]
        self.lbw += lb
        self.ubw += ub
        self.w0 += w0
        if isinstance(discrete, List):
            if len(discrete) != len(lb):
                raise Exception("Discrete list doesn't have the same size as lb!")
            self.discrete += discrete
        else:
            self.discrete += [1 if discrete else 0] * len(lb)
        return idx

    def sym(
        self,
        name: str,
        shape: Union[int, Tuple[int, int]],
        lb: Union[float, List[float]],
        ub: Union[float, List[float]],
        w0: Optional[Union[float, List[float]]] = None,
        discrete: Union[bool, List[int]] = False,
    ) -> GlobalSettings.CASADI_VAR:
        """Create a symbolic variable."""
        nr = count_shape(shape)

        # Gather Data
        if name not in self.indices:
            self.indices[name] = []
        idx_list = self.indices[name]

        # Create
        x = GlobalSettings.CASADI_VAR.sym("%s[%d]" % (name, len(idx_list)), nr)
        lb = make_list(lb, nr)
        ub = make_list(ub, nr)

        if w0 is None:
            w0 = lb
        w0 = make_list(w0, nr)

        if len(lb) != nr:
            raise Exception("Lower bound error!")
        if len(ub) != nr:
            raise Exception("Upper bound error!")
        if len(w0) != nr:
            raise Exception("Estimation length error (w0 %d vs nr %d)!" % (len(w0), nr))

        # Collect
        out = self.add_w(lb, x, ub, w0, discrete)
        idx_list.append(out)
        return as_shape(x, shape)

    def add_parameters(self, name, shape: Union[int, Tuple[int, int]], values=0):
        """Add some parameters."""
        nr = count_shape(shape)
        # Gather Data
        if name not in self.indices_p:
            self.indices_p[name] = []
        idx_list = self.indices_p[name]

        # Create
        p = GlobalSettings.CASADI_VAR.sym("%s[%d]" % (name, len(idx_list)), nr)
        values = make_list(values, nr)

        # Create & Collect
        new_idx = [i + len(self.p0) for i in range(nr)]
        self.p += [p]
        self.p0 += values

        self.indices_p[name].extend(new_idx)
        return as_shape(p, shape)

    def add_g(
        self,
        mini: float,
        equation: GlobalSettings.CASADI_VAR,
        maxi: float,
        is_linear=0,
        is_discrete=0,
        is_dwell_time=0,
    ) -> int:
        """
        Add to g.

        :param mini: minimum
        :param equation: constraint equation
        :param maxi: maximum
        :return: index of constraint
        """
        if self.check_illconditioned:
            if (abs(maxi) > 1e5 and not isinf(maxi)) or (
                abs(mini) > 1e5 and not isinf(mini)
            ):
                raise Exception(
                    "Illconditioned {mini} <= {eq} <= {maxi}".format(
                        mini=mini, eq=equation, maxi=maxi
                    )
                )

        nr = equation.shape[0] * equation.shape[1]
        equation = reshape(equation, (nr, 1))
        self.lbg += make_list(mini, nr)
        self.g += make_list(equation)
        self.ubg += make_list(maxi, nr)
        self.g_lin += make_list(int(is_linear), nr)
        self.g_dis += make_list(int(is_discrete), nr)
        self.g_dwelltime += make_list(int(is_dwell_time), nr)
        return len(self.ubg) - 1

    def add_soc(self, equation: GlobalSettings.CASADI_VAR):
        self.h = make_soc_matrix(self.h, equation)

    def leq(self, op1, op2, is_linear=0, is_discrete=0, is_dwell_time=0):
        """Lower or equal."""
        if isinstance(op1, (float, int, list)):
            self.add_g(
                op1,
                op2,
                inf,
                is_linear=is_linear,
                is_discrete=is_discrete,
                is_dwell_time=is_dwell_time,
            )
        elif isinstance(op2, (float, int, list, np.ndarray)):
            self.add_g(
                -inf,
                op1,
                op2,
                is_linear=is_linear,
                is_discrete=is_discrete,
                is_dwell_time=is_dwell_time,
            )
        else:
            diff = op1 - op2
            self.add_g(
                -inf,
                diff,
                0,
                is_linear=is_linear,
                is_discrete=is_discrete,
                is_dwell_time=is_dwell_time,
            )

    def eq(self, op1, op2, is_linear=0, is_discrete=0):
        """Equal."""
        self.add_g(0, op1 - op2, 0, is_linear=is_linear, is_discrete=is_discrete)

    def sym_bool(self, name: str, nr: int = 1, w0=1) -> GlobalSettings.CASADI_VAR:
        """Create a symbolic boolean."""
        return self.sym(name, nr, 0, 1, w0=w0, discrete=True)

    def get_gauss_newton_hessian(self) -> GlobalSettings.CASADI_VAR:
        x = vertcat(*self.w)
        if isinstance(self.r, list):
            self.r = vertcat(*self.r)

        dr = jacobian(self.r, x)
        return dr.T @ dr

    def create_f(self):
        """Create function f."""
        return Function("f", [vcat(self.w), vcat(self.p)], [self.f])

    def create_g(self):
        """Create function f."""
        return Function("g", [vcat(self.w), vcat(self.p)], [vcat(self.g)])

    def _r_exist(self):
        return not (isinstance(self.r, list) and len(self.r) == 0)

    def get_problem(self, with_gn=True) -> MinlpProblem:
        """Extract problem."""
        idx_x_integer = [i for i, v in enumerate(self.discrete) if v == 1]
        idx_g_dwelltime = [i for i, elm in enumerate(self.g_dwelltime) if elm == 1]
        idx_g_without_dwelltime = [
            i for i, elm in enumerate(self.g_dwelltime) if elm == 0
        ]
        if idx_g_dwelltime == []:
            idx_g_dwelltime = None
        if idx_g_without_dwelltime == []:
            idx_g_without_dwelltime = None
        if self._r_exist() and with_gn:
            gn_hessian = self.get_gauss_newton_hessian()
        else:
            gn_hessian = None

        return MinlpProblem(
            f=self.f,
            g=vertcat(*self.g),
            h=self.h,
            x=vertcat(*self.w),
            idx_x_integer=idx_x_integer,
            p=vertcat(*self.p),
            gn_hessian=gn_hessian,
            idx_g_dwelltime=idx_g_dwelltime,
            idx_g_without_dwelltime=idx_g_without_dwelltime,
        )

    def get_data(self) -> MinlpData:
        """Get data structure."""
        return MinlpData(
            p=self.p0,
            x0=DM(self.w0),
            _lbx=DM(self.lbw),
            _ubx=DM(self.ubw),
            _lbg=DM(self.lbg),
            _ubg=DM(self.ubg),
        )

    def get_indices(self, name: str, is_param: bool = False):
        """
        Get indices of a certain variable.

        :param name: name
        """
        if is_param:
            return self.indices_p[name]
        else:
            return self.indices[name]

    def check(self):
        """Test properties."""
        nr_of_w = vertcat(*(self.w)).size()[0]
        if nr_of_w != len(self.lbw):
            raise Exception(
                "Length lbw (%d) incorrect, need %d." % (len(self.lbw), nr_of_w)
            )
        if nr_of_w != len(self.ubw):
            raise Exception(
                "Length ubw (%d) incorrect, need %d." % (len(self.ubw), nr_of_w)
            )

        nr_of_g = vertcat(*(self.g)).size()[0]
        if nr_of_g != len(self.lbg):
            raise Exception(
                "Length lbg (%d) incorrect, need %d." % (len(self.lbg), nr_of_w)
            )
        if nr_of_g != len(self.ubg):
            raise Exception(
                "Length ubg (%d) incorrect, need %d." % (len(self.ubg), nr_of_w)
            )
        print("Nr of w %d - Nr of g %d" % (nr_of_w, nr_of_g))

    def check_hessian(self, plot=True, tolerance=1e-8):
        """Check hessian, returns the hessian h, jacobian j and eigenvalues."""
        import numpy as np

        w = vertcat(*self.w)
        h = hessian(self.f, w)
        h_func = Function("h", [w], [h[0]])
        h_val = h_func(np.ones((w.size(1), w.size(2))))
        eig_values = np.unique(np.linalg.eig(h_val)[0])
        print(f"Eigenvalues: {eig_values}")
        if plot:
            import matplotlib.pyplot as plt

            plt.figure()
            plt.subplot(1, 2, 1)
            plt.spy(h_val)
            plt.subplot(1, 2, 2)
            plt.imshow(h_val, interpolation="none", cmap="binary")
            plt.colorbar()
        if (eig_values < -tolerance).any():
            print("Hessian is not semi-definite!")
        return h[0], h[1], eig_values



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Validation tools.

Several validation tools to check if a solution is feasible and valid.
"""

from camino.problem import MinlpProblem
from camino.data import MinlpData
import casadi as ca
from camino.settings import Settings
import numpy as np
from camino.utils.conversion import to_float


def check_solution(
    problem: MinlpProblem,
    data: MinlpData,
    x_star,
    s: Settings,
    throws=True,
    check_objval=True,
):
    """Check a solution."""
    f = ca.Function("f", [problem.x, problem.p], [problem.f])
    g = ca.Function("g", [problem.x, problem.p], [problem.g])
    f_val = to_float(f(x_star, data.p).full())
    g_val = g(x_star, data.p).full().squeeze()
    lbg, ubg = data.lbg.squeeze(), data.ubg.squeeze()
    print(f"Objective value {f_val} (real) vs {data.obj_val}")
    msg = []
    if check_objval and abs(to_float(data.obj_val) - f_val) > s.EPS:
        msg.append(
            f"Objective value wrong, with error {abs(to_float(data.obj_val) - f_val) - s.EPS}"
        )
    if np.any(data.lbx > x_star + s.EPS):
        msg.append(f"Lbx > x* for indices:\n{np.nonzero(data.lbx > x_star + s.EPS).T}")
    if np.any(data.ubx < x_star - s.EPS):
        msg.append(f"Ubx > x* for indices:\n{np.nonzero(data.ubx < x_star - s.EPS).T}")
    if np.any(lbg > g_val + s.EPS):
        msg.append(
            "Lbg > g(x*,p) + tol:\n"
            f"{(lbg - g_val - s.EPS)[np.nonzero(lbg - g_val - s.EPS >0)]}"
        )
        msg.append("for indices:\n" f"{np.nonzero(lbg - g_val - s.EPS >0)}")
    if np.any(ubg < g_val - s.EPS):
        msg.append(
            "Ubg < g(x*,p) - tol:\n"
            f"{(g_val - ubg - s.EPS)[np.nonzero(g_val - ubg - s.EPS > 0)]}"
        )
        msg.append("for indices:\n" f"{np.nonzero(g_val - ubg - s.EPS > 0)}")
    check_integer_feasible(problem.idx_x_integer, x_star, s, throws=throws)
    if msg:
        msg = "\n".join(msg)
        if throws:
            raise Exception(msg)
        else:
            print(msg)


def check_integer_feasible(idx_x_integer, x_star, s: Settings, throws=True):
    """Check if the solution is integer feasible."""
    x_bin = np.array(x_star)[idx_x_integer].squeeze()
    x_bin_rounded = np.round(x_bin)
    if np.any(np.abs(x_bin_rounded - x_bin) > s.EPS):
        idx = np.nonzero(np.abs(x_bin_rounded - x_bin) > s.CONSTRAINT_INT_TOL)
        msg = f"Integer infeasible: {x_bin[idx]} {idx}"
        if throws:
            raise Exception(msg)
        else:
            print(throws)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Car model.

Based on car model of Massimo De Mauri
"""

import csv
import casadi as ca
import numpy as np
import matplotlib.pyplot as plt
from os import path

from typing import Union
from camino.settings import GlobalSettings
from camino.problems import MinlpData, MetaDataMpc, MinlpProblem
from camino.problems.gearbox.models import (
    insight_50kw_power_jc,
    advisor_em_pwr_sc,
    battery_hu,
)
from camino.problems.dsc import Description
from camino.utils.cache import CachedFunction


def get_cycle_data(dt=1, N=30):
    """Get cycle data."""
    current_directory = path.abspath(path.dirname(__file__))
    # import the driving cycle
    with open(current_directory + "/cycle.csv", "r") as csvfile:
        cycle_dat = [float(s) for s in list(csv.reader(csvfile, delimiter=","))[0]][20:]

    xgrid = np.linspace(0, len(cycle_dat), len(cycle_dat))
    f = ca.interpolant("data", "bspline", [xgrid], cycle_dat)
    cycle = f(np.arange(0, dt * (N + 1), dt))
    return cycle, cycle[1:] - cycle[:-1]


def create_gear_step(dsc, nr_gears, gears_prev, gears_prev2, allow_stop=False):
    """Create a setp of gears."""
    gears = dsc.sym_bool("gears", nr_gears)
    dsc.eq(ca.sum1(gears), 1)

    if allow_stop:
        dsc.leq(gears[0], gears_prev[0] + gears_prev[1])
        dsc.leq(gears[0], gears_prev2[0] + gears_prev2[1])

    for j in range(1, nr_gears - 1):
        dsc.leq(gears[j], gears_prev[j - 1] + gears_prev[j] + gears_prev[j + 1])
        dsc.leq(gears[j], gears_prev2[j - 1] + gears_prev2[j] + gears_prev2[j + 1])

    j = nr_gears - 1
    dsc.leq(gears[j], gears_prev[j - 1] + gears_prev[j])
    dsc.leq(gears[j], gears_prev2[j - 1] + gears_prev2[j])
    return gears


def create_gearbox_int(N=10, traject=None):
    """Create integer gearbox."""
    if traject is None:
        traject = [np.sin(i / 10) * 4 + 5 for i in range(N)]

    def cost_gear(j, setpoint):
        """Cost to run at setpoint for gear j."""
        return (10 * (setpoint - j * 2 + 1)) ** 2 + 5 * setpoint + j

    dsc = Description()

    traject = dsc.add_parameters("traject", N, traject)
    nr_gears = 5
    gears_prev2 = dsc.add_parameters("gears-1", 1, [0])
    gears_prev = dsc.add_parameters("gears0", 1, [0])
    throttle_prev = dsc.add_parameters("throttle", 1, [0])
    for i in range(N):
        gears = dsc.sym("gear", 1, 0, 5, 0, discrete=True)
        throttle = dsc.sym("throttle", 1, 0, 10)
        dsc.leq(gears - gears_prev, 1)
        dsc.leq(gears_prev - gears, 1)
        dsc.leq(gears - gears_prev2, 1)
        dsc.leq(gears_prev2 - gears, 1)
        dsc.leq(throttle - throttle_prev, 0.5)
        dsc.leq(throttle_prev - throttle, 0.5)

        dsc.leq(throttle, gears + 1)
        dsc.leq(gears - 1, throttle)

        for j in range(nr_gears):
            dsc.f += (
                (traject[i] - throttle) ** 4 + (traject[i] - throttle) ** 2 + gears * 10
            )

        # swcost = dsc.sym("swcost", 1, 0, 1)
        # dsc.f += swcost
        # dsc.leq(gears - gears_prev, swcost)

        gears_prev2 = gears_prev
        gears_prev = gears
        throttle_prev = throttle

    def plot_gearbox(data: MinlpData, x_star):
        """Plot gearbox."""
        gear_idcs = dsc.get_indices("gear")
        gears = [float(x_star[gear_idx]) for gear_idx in gear_idcs]
        plt.scatter(list(range(len(gears))), gears)
        plt.show()

    problem = dsc.get_problem()
    problem.meta = MetaDataMpc()
    problem.meta.plot = plot_gearbox
    problem.meta.shift = plot_gearbox

    return problem, dsc.get_data()


def create_simple_gearbox(N=10, traject=None):
    """Very simple gearbox."""
    if traject is None:
        traject = [np.sin(i / 10) * 4 + 5 for i in range(N)]

    def cost_gear(j, setpoint):
        """Cost to run at setpoint for gear j."""
        return (10 * (setpoint - j * 2 + 1)) ** 2 + 5 * setpoint + j

    dsc = Description()

    traject = dsc.add_parameters("traject", N, traject)
    nr_gears = 5
    gears_prev2 = dsc.add_parameters("gears-1", nr_gears, [1, 0, 0, 0, 0])
    gears_prev = dsc.add_parameters("gears0", nr_gears, [1, 0, 0, 0, 0])
    for i in range(N):
        gears = create_gear_step(dsc, nr_gears, gears_prev, gears_prev2)
        for j in range(nr_gears):
            dsc.f += gears[j] * cost_gear(j, traject[i])

        swcost = dsc.sym("swcost", 1, 0, 1)
        dsc.f += swcost
        for j in range(1, nr_gears):
            dsc.leq(gears[j] - gears_prev[j - 1], swcost)

        gears_prev2 = gears_prev
        gears_prev = gears

    def plot_gearbox(data: MinlpData, x_star):
        """Plot gearbox."""
        gear_idcs = dsc.get_indices("gears")
        gears = [
            float(sum([x_star[k] * (1 + i) for i, k in enumerate(gear_idx)]))
            for gear_idx in gear_idcs
        ]
        plt.scatter(list(range(len(gears))), gears)
        plt.show()

    problem = dsc.get_problem()
    problem.meta = MetaDataMpc()
    problem.meta.plot = plot_gearbox
    problem.meta.shift = plot_gearbox

    return problem, dsc.get_data()


def create_gearbox(
    N=10, gearbox_type="gasoline", dt=1.0, switch_cost=False
) -> Union[MinlpProblem, MinlpData]:
    """Create a gearbox model."""
    N = int(N)
    cycle, dcycle = get_cycle_data(dt, N)

    # ------------------------------------------------
    # Model
    # ------------------------------------------------
    base_mass = 800
    wheelR = 0.285
    drag_area = 2
    air_density = 1.225
    drag_coeff = 0.35

    # temporary variables
    w = GlobalSettings.CASADI_VAR.sym("w")
    P = GlobalSettings.CASADI_VAR.sym("P")
    Prat = GlobalSettings.CASADI_VAR.sym("Prat")
    Erat = GlobalSettings.CASADI_VAR.sym("Erat")
    SoC = GlobalSettings.CASADI_VAR.sym("SoC")

    # ICE
    ICE_wrat = 6000 * np.pi / 30
    ICE_data = insight_50kw_power_jc()
    ICE_dFin = ca.Function(
        "ICE_dFin", [Prat, w, P], [ICE_data["dFin"](ICE_wrat, 1000 * Prat, w, 1000 * P)]
    )
    ICE_Fstart = ca.Function("ICE_Fstart", [Prat], [ICE_data["Fstart"](1000 * Prat)])
    ICE_Pmax = ca.Function(
        "ICE_Pmax", [Prat, w], [ICE_data["Pmax"](ICE_wrat, 1000 * Prat, w) * 1e-3]
    )
    ICE_mass = ca.Function("ICE_mass", [Prat], [ICE_data["mass"](1000 * Prat)])
    ICE_minw = max(ICE_data["minw"](ICE_wrat), 1000 * np.pi / 30)

    # EM
    EM_wrat = 10000 * np.pi / 30
    EM_data = advisor_em_pwr_sc()
    EM_Pin = ca.Function(
        "EM_Pin",
        [Prat, w, P],
        [EM_data["Pin"](EM_wrat, 1000 * Prat, w, 1000 * P) * 1e-3],
    )
    # EM_Pmax = ca.Function('EM_Pmax', [Prat, w], [
    #                       EM_data['Pmax'](EM_wrat, 1000 * Prat, w) * 1e-3])
    EM_mass = ca.Function("EM_mass", [Prat], [EM_data["mass"](1000 * Prat)])

    # battery
    BT_data = battery_hu()
    SoC_max = BT_data["SoC_max"]
    SoC_min = BT_data["SoC_min"]
    BT_dSoC = ca.Function(
        "BT_dSoC", [Erat, SoC, P], [BT_data["dSoC"](Erat * 3.6e6, SoC, 1000 * P)]
    )
    BT_Pout = ca.Function(
        "BT_Pout", [Erat, SoC, P], [BT_data["Pout"](Erat * 3.6e6, SoC, 1000 * P) * 1e-3]
    )
    BT_Pmax = ca.Function(
        "BT_Pmax", [Erat, SoC], [BT_data["Pmax"](Erat * 3.6e6, SoC) * 1e-3]
    )
    BT_Pmin = ca.Function(
        "BT_Pmin", [Erat, SoC], [BT_data["Pmin"](Erat * 3.6e6, SoC) * 1e-3]
    )
    BT_mass = ca.Function("BT_mass", [Erat], [BT_data["mass"](Erat * 3.6e6)])

    # ------------------------------------------------
    # Parameters initial values
    # ------------------------------------------------

    # speeds to guess ratios
    reference_speeds = [15, 30, 55, 85, 115]
    # reference_speeds = [30, 55, 85, 115, 155]

    # for diesel:
    referece_ICEspeed_diesel = 2000
    R_diesel = [
        (referece_ICEspeed_diesel * np.pi / 30) / ((reference_speeds[k] / 3.6) / wheelR)
        for k in range(len(reference_speeds))
    ]

    # for gasoline
    referece_ICEspeed_gasoline = 2500
    R_gasoline = [
        (referece_ICEspeed_gasoline * np.pi / 30)
        / ((reference_speeds[k] / 3.6) / wheelR)
        for k in range(len(reference_speeds))
    ]

    # # from advisor default parallel
    # Rem_advisor = 0.99 * (EM_wrat / ICE_wrat)
    # R_advisor = [13.33, 7.57, 5.01, 3.77, 2]

    # ------------------------------------------------
    # Problem definition
    # ------------------------------------------------

    # fixed parameters
    if gearbox_type == "diesel":
        R = R_diesel
    else:
        R = R_gasoline

    Rem = (3000 * np.pi / 30) / ((60 / 3.6) / wheelR)
    ICE_Prat = 55
    EM_Prat = 25
    BT_Erat = 1.024  # kWh
    TNK_Frat = 30000
    TNK_Finit = TNK_Frat / 2
    # SoC_opt = SoC_min + 0.75 * (SoC_max - SoC_min)
    SoC_start = SoC_min + 0.1 * (SoC_max - SoC_min)

    dsc = Description()
    SoC = dsc.sym("SoC", N + 1, float(SoC_min), float(SoC_max), float(SoC_start))
    dsc.eq(float(SoC_start), SoC[0])
    Fuel = dsc.sym("Fuel", N + 1, 0, float(TNK_Frat), float(TNK_Finit))
    dsc.add_g(float(TNK_Finit), Fuel[0], float(TNK_Finit))

    # # Switch on/off
    # # Create gearbox
    cost_switch = 0.4 * float(ICE_Fstart(ICE_Prat))

    # cost_switch = 0  # 0.4 * ICE_Fstart(ICE_Prat)
    # # float(0.4 * 45.6 * ICE_Fstart(ICE_Prat))

    # Continous variables
    dFuel_max = 20.0
    dFuel = dsc.sym("dFuel", N, 0, dFuel_max, 0)
    Pb = dsc.sym("Pb", N, float(BT_Pmin(BT_Erat, 1.0)), float(BT_Pmax(BT_Erat, 1.0)), 0)
    Pice = dsc.sym("Pice", N, 0, ICE_Prat, 0)
    Pem = dsc.sym("Pem", N, -EM_Prat, EM_Prat, 0)

    nr_gears = len(R) + 1
    gears_init = [0] * nr_gears
    gears_init[0] = 1
    gearbox_prev = dsc.add_parameters("gears0", nr_gears, gears_init)
    gearbox_prev2 = dsc.add_parameters("gears-1", nr_gears, gears_init)

    # precalculations
    mass = base_mass + ICE_mass(ICE_Prat) + EM_mass(EM_Prat) + BT_mass(BT_Erat)

    for i in range(N):
        gearbox = create_gear_step(dsc, nr_gears, gearbox_prev, gearbox_prev2)
        if switch_cost:
            dsc.f += ca.sum1((gearbox_prev - gearbox) ** 2 * cost_switch / 2)
        gearbox_prev2 = gearbox_prev
        gearbox_prev = gearbox

        # gearbox = create_gear_step(dsc, nr_gears, gears_prev, gears_prev, allow_stop=True)
        off = gearbox[0]
        gears = gearbox[1:]

        Freq = (
            mass * dcycle[i]
            + 0.5 * air_density * drag_coeff * drag_area * cycle[i] ** 2
        )
        # Treq = Freq * wheelR
        Paux = 0.3  # accessory power load kW
        Preq = Freq * cycle[i] / 1000
        wem = Rem * cycle[i] / wheelR
        wice = sum([gears[k] * R[k] * cycle[i] / wheelR for k in range(len(R))])

        # # ODE
        dsc.eq(BT_dSoC(BT_Erat, SoC[i], Pb[i]) + (SoC[i] - SoC[i + 1]) / dt, 0)
        dsc.eq((Fuel[i] - Fuel[i + 1]) / dt - dFuel[i], 0)
        # max_fuel * (1- off) > dFuel)
        dsc.leq(dFuel[i] - (1 - off) * dFuel_max, 0)

        # EQ
        # Power required = P_em + P_ice
        dsc.leq(Preq - Pem[i] - Pice[i], 0)

        # wice < ICE_wrat
        dsc.leq(wice, ICE_wrat * 0.5)
        dsc.leq(ICE_minw * (1 - off) - wice, 0)
        dsc.leq(Pice[i] - ICE_Prat * (1 - off), 0)
        for k in range(ICE_Pmax(0, 0).numel()):
            dsc.leq(Pice[i] - ICE_Pmax(ICE_Prat, wice)[k], 0)

        for k in range(ICE_dFin(0, 0, 0).numel()):
            dsc.leq(
                (
                    ICE_dFin(ICE_Prat, wice, Pice[i])[k]
                    - off * ICE_dFin(ICE_Prat, 0, 0)[k]
                    - dFuel[i]
                ),
                0,
            )

        # Following equations are OK!
        dsc.leq(Pb[i] - BT_Pmax(BT_Erat, SoC[i]), 0)
        dsc.leq(0, Pb[i] - BT_Pmin(BT_Erat, SoC[i]))
        for k in range(EM_Pin(0, 0, 0).numel()):
            dsc.leq(
                (
                    EM_Pin(EM_Prat, wem, Pem[i])[k]
                    + Paux
                    - BT_Pout(BT_Erat, SoC[i], Pb[i])
                ),
                0,
            )
        dsc.add_g(-EM_Prat, Pem[i], EM_Prat)

        # Pb = battery consumption (Power battery)
        dsc.f += 0.4 * 45.6 * dFuel[i] + Pb[i]

    dsc.f += 200 * (
        -ca.log((SoC[-1] - SoC_min) / (SoC_max - SoC_min))
        - ca.log((SoC_max - SoC[-1]) / (SoC_max - SoC_min))
        + 2 * ca.log(0.5)
    )

    def plot_gearbox(data: MinlpData, x_star):
        """Plot gearbox."""
        # from camino.utils import latexify
        # latexify()

        t = np.arange(0, (N + 1) * dt, dt)
        nr_plots = 6
        fig, axs = plt.subplots(nr_plots, 1, figsize=(6, 8))

        axs[0].plot(t, cycle, label="Traject")
        plot_names = [["SoC"], ["Fuel"], ["dFuel"], ["Pb", "Pice", "Pem"]]
        for i, ax in enumerate(axs[1:-1]):
            for name in plot_names[i]:
                y_data = x_star[dsc.get_indices(name)[0]]
                ax.plot(t[: y_data.numel()], y_data, label=name)

        gear_idcs = dsc.get_indices("gears")
        gears = [
            float(sum([x_star[k] * i for i, k in enumerate(gear_idx)]))
            for gear_idx in gear_idcs
        ]

        axs[-1].scatter(t[:-1], gears, label="gear shift")
        for i in range(nr_gears):
            axs[-1].plot([0, N * dt], [i, i], ":", color="tab:grey")
        axs[-1].set_ylim([0, nr_gears])

        for ax in axs:
            ax.set_xlim(0, (N + 1) * dt)
            ax.legend(loc="upper right", fontsize="small").set_draggable(True)

        fig.subplots_adjust(hspace=0.5)
        plt.show()

    dsc.w = [ca.vcat(dsc.w)]
    dsc.p = [ca.vcat(dsc.p)]
    dsc.f = CachedFunction(f"gearbox{N}", dsc.create_f)(dsc.w[0], dsc.p[0])
    dsc.g = [CachedFunction(f"gearbox{N}", dsc.create_g)(dsc.w[0], dsc.p[0])]

    problem = dsc.get_problem()
    problem.meta = MetaDataMpc()
    problem.meta.plot = plot_gearbox
    problem.meta.shift = plot_gearbox

    return problem, dsc.get_data()



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Statistics class."""

from copy import deepcopy
from datetime import datetime
from dataclasses import dataclass, field
import os
from typing import Dict, List, Optional
from camino.utils import toc
from camino.utils.conversion import to_0d
from camino.settings import GlobalSettings
from camino.utils.data import save_pickle
from camino.data import MinlpData


@dataclass()
class Stats:
    """
    Collecting statistics


    There are some shared properties between the algorithms:
        iter_nr: Number of iterations performed
        lb: Lower bound value
        ub: Upper bound value
        relaxed: Relaxed solution
    """

    mode: str
    problem_name: str
    nr_reset: int = 0
    datetime: str = field(
        default_factory=lambda: datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
    )
    data: Dict[str, float] = field(default_factory=lambda: {})
    out_dir: str = GlobalSettings.OUT_DIR
    _full_stats_to_pickle: List = field(default_factory=lambda: [])
    _relaxed: Optional[MinlpData] = None

    def __getitem__(self, key):
        """Get attribute."""
        if key not in self.data:
            return 0
        return self.data[key]

    def __setitem__(self, key, value):
        """Set item."""
        self.data[key] = value

    @property
    def relaxed_solution(self):
        """Relaxed value."""
        return deepcopy(self._relaxed)

    @relaxed_solution.setter
    def relaxed_solution(self, value):
        """Set relaxed solution value."""
        self._relaxed = deepcopy(value)

    def print(self):
        """Print statistics."""
        print("Statistics")
        for k, v in sorted(self.data.items()):
            if k not in [
                "iterate_data",
                "solutions_all",
                "solved_all",
                "solutions",
                "mip_solutions_all",
                "mip_solved_all",
            ]:
                if k == "sol_x":
                    if len(self.data["sol_x"]) < 10:
                        print(f"\t{k}: {v}")
                    else:
                        print(f"\t{k}: {v[:5]}, ...")
                else:
                    print(f"\t{k}: {v}")

    def reset(self):
        """Reset the statistics."""
        self.data = {}
        self._full_stats_to_pickle = {}
        self.nr_reset += 1

    def save(self, dest=None):
        """Save statistics."""
        time = toc()  # TODO add time
        if dest is None:
            dest = os.path.join(
                self.out_dir,
                f"{self.datetime}_{self.mode}_{self.problem_name}_{self.nr_reset}.pkl",
            )
        print(f"Saving to {dest}")
        data = self.data.copy()
        to_pickle = []
        general_stats = {}
        for key, value in data.items():
            if key not in [
                "solutions_all",
                "solved_all",
                "mip_solutions_all",
                "mip_solved_all",
            ]:
                general_stats[key] = value
        general_stats["time"] = time
        try:
            for idx, (elm, mip_elm) in enumerate(
                zip(data["solutions_all"], data["mip_solutions_all"])
            ):
                tmp_dict = {}
                tmp_dict.update(general_stats)
                tmp_dict["sol_pool_idx"] = idx
                tmp_dict["sol_pool_success"] = data["solved_all"][idx]
                tmp_dict["sol_pool_objective"] = float(elm["f"])
                tmp_dict["sol_pool_x"] = to_0d(elm["x"])
                tmp_dict["mip_sol_pool_idx"] = idx
                tmp_dict["mip_sol_pool_success"] = data["mip_solved_all"][idx]
                tmp_dict["mip_sol_pool_objective"] = float(mip_elm["f"])
                tmp_dict["mip_sol_pool_x"] = to_0d(mip_elm["x"])
                to_pickle.append(tmp_dict)
        except Exception:
            tmp_dict = {}
            tmp_dict.update(general_stats)
            to_pickle.append(tmp_dict)

        self._full_stats_to_pickle += to_pickle
        save_pickle(self._full_stats_to_pickle, dest)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Cached function."""

import logging
from os import path, rename
from typing import Callable
import casadi as ca
from subprocess import call

from camino.settings import GlobalSettings
from camino.utils.data import read_json, write_json, load_pickle, save_pickle

logger = logging.getLogger(__name__)
_COMPILERS = ["gcc-15", "gcc"]
_COMPILER = None


def get_compiler():
    """Get available compiler."""
    global _COMPILER
    if _COMPILER is None:
        for compiler in _COMPILERS:
            try:
                call([compiler, "--version"])
                _COMPILER = compiler
                break
            except Exception:
                pass

    return _COMPILER


def compile(input_file, output_file, options=None):
    """Compile a c file to an so file."""
    compiler = get_compiler()

    _CXX_FLAGS = ["-fPIC", "-v", "-shared", "-fno-omit-frame-pointer", "-O2"]
    call([compiler] + _CXX_FLAGS + ["-o", output_file, input_file])


def cache_data(name, generator_func, *args, **kwargs):
    """Cache data."""
    name = name + "_" + getattr(generator_func, "__name__", "Unknown")
    filename = path.join(GlobalSettings.CACHE_FOLDER, name + ".pkl")
    if path.exists(filename):
        return load_pickle(filename)
    else:
        data = generator_func(*args, **kwargs)
        save_pickle(data, filename)
        return data


def return_func(func):
    """Create a function."""

    def r():
        return func

    return r


class CachedFunction:
    """A cached function."""

    def __init__(
        self, name, func: Callable[[], ca.Function], filename=None, do_compile=None
    ):
        """Load or create a cached function."""
        if do_compile is None:
            do_compile = True

        self.name = name + "_" + getattr(func, "__name__", "Unknown")
        if filename is None:
            self.filename = path.join(GlobalSettings.CACHE_FOLDER, self.name)
        else:
            self.filename = filename

        if self._exist_so():
            logger.debug(f"Loading function from so-file {self.name}")
            self._load_so()
        elif self._exist():
            logger.debug(f"Loading function from disc {self.name}")
            self._load()
        else:
            logger.debug(f"Creating function {self.name}")
            self._create(func)
            if do_compile:
                logger.debug(f"Compiling function {self.name}")
                self._save_so()
                self._load_so()
            else:
                self._save()

    def _create(self, func):
        """Create a function."""
        self.f = func()

    def _exist(self):
        """Check if the file exist."""
        return path.exists(self.filename + ".cache")

    def _save(self):
        """Save to a file."""
        self.f.save(self.filename + ".cache")

    def _load(self):
        """Load the function."""
        self.f = ca.Function.load(self.filename + ".cache")

    def _exist_so(self):
        """Check if the file exist."""
        return path.exists(self.filename + ".json")

    def _save_so(self):
        """Save as so file."""
        data = {
            "lib_file": self.filename + ".so",
            "c_file": self.filename + ".c",
            "func_name": self.f.name(),
        }

        tmp_file = data["func_name"] + ".c"
        cg = ca.CodeGenerator(tmp_file)
        cg.add(self.f)
        cg.add(self.f.jacobian())
        cg.add(self.f.jacobian().jacobian())
        cg.generate()
        rename(tmp_file, data["c_file"])
        compile(data["c_file"], data["lib_file"])
        write_json(data, self.filename + ".json")

    def _load_so(self):
        """Load an SO file."""
        data = read_json(self.filename + ".json")
        self.f = ca.external(data["func_name"], data["lib_file"])

    def __call__(self, *args, **kwargs):
        """Call the function."""
        return self.f(*args, **kwargs)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""An feasibility pump implementation."""

import numpy as np
import casadi as ca
from camino.solvers import (
    SolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import GlobalSettings, Settings
from camino.utils import logging
from camino.utils.conversion import to_0d

logger = logging.getLogger(__name__)


class LinearProjection(SolverClass):
    """
    Create NLP solver.

    This solver solves an NLP problem. This is either relaxed or the binaries are fixed.  The NLP aims to satisfy
    original constraint in g and minimize the distance of relaxed x_bin from their rounded value in L1-norm. The
    rounding procedure follows a rule.
    """

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        super(LinearProjection, self).__init__(problem, stats, s)
        options = regularize_options(
            s.IPOPT_SETTINGS, {"jit": s.WITH_JIT, "ipopt.max_iter": 5000}, s
        )

        self.idx_x_integer = problem.idx_x_integer
        x_bin_var = problem.x[self.idx_x_integer]
        self.nr_x_bin = x_bin_var.shape[0]
        rounded_value = GlobalSettings.CASADI_VAR.sym("rounded_value", self.nr_x_bin)
        slack_variables = GlobalSettings.CASADI_VAR.sym("slack", self.nr_x_bin)

        penalty_term = ca.sum1(slack_variables)
        g = ca.vertcat(
            problem.g,
            # x_bin_var - rounded < slack
            # - (x_bin_var - rounded) < slack
            slack_variables - (x_bin_var - rounded_value),
            slack_variables + (x_bin_var - rounded_value),
        )
        self.solver = ca.nlpsol(
            "nlpsol",
            "ipopt",
            {
                "f": penalty_term,
                "g": g,
                "x": ca.vertcat(problem.x, slack_variables),
                "p": ca.vertcat(problem.p, rounded_value),
            },
            options,
        )

    def solve(self, nlpdata: MinlpData, **kwargs) -> MinlpData:
        success_out = []
        sols_out = []
        for sol in nlpdata.solutions_all:
            lbx = nlpdata.lbx
            ubx = nlpdata.ubx

            x_bin_var = to_0d(sol["x"][self.idx_x_integer])
            new_sol = self.solver(
                x0=ca.vertcat(nlpdata.x0, np.zeros(self.nr_x_bin)),
                lbx=ca.vertcat(lbx, np.zeros(self.nr_x_bin)),
                ubx=ca.vertcat(ubx, ca.inf * np.ones(self.nr_x_bin)),
                lbg=ca.vertcat(nlpdata.lbg, np.zeros(2 * self.nr_x_bin)),
                ubg=ca.vertcat(nlpdata.ubg, ca.inf * np.ones(2 * self.nr_x_bin)),
                p=ca.vertcat(nlpdata.p, x_bin_var),
            )

            success, _ = self.collect_stats("FP", sol=new_sol)
            if not success:
                logger.debug("Infeasible solution!")
            success_out.append(success)
            sols_out.append(new_sol)

        nlpdata.prev_solutions = sols_out
        nlpdata.solved_all = success_out
        return nlpdata


class ObjectiveLinearProjection(SolverClass):
    """
    Create NLP solver.

    This solver solves an NLP problem. This is either relaxed or the binaries are fixed. The NLP aims to satisfy
    original constraint in g and minimize a linear combination of the original objective and the distance of relaxed
    x_bin from their rounded value in L1-norm. The rounding procedure follows a rule and the weight of the L1-norm
    term in the objective is increased during iterations.
    """

    def __init__(self, problem: MinlpProblem, stats: Stats, s: Settings):
        super(ObjectiveLinearProjection, self).__init__(problem, stats, s)
        options = regularize_options(s.IPOPT_SETTINGS, {}, s)

        self.idx_x_integer = problem.idx_x_integer
        x_bin_var = problem.x[self.idx_x_integer]
        self.nr_x_bin = x_bin_var.shape[0]
        rounded_value = GlobalSettings.CASADI_VAR.sym("rounded_value", self.nr_x_bin)
        slack_variables = GlobalSettings.CASADI_VAR.sym("slack", self.nr_x_bin)
        alpha = GlobalSettings.CASADI_VAR.sym("alpha", 1)
        int_error = GlobalSettings.CASADI_VAR.sym("int_error", 1)
        obj_val = GlobalSettings.CASADI_VAR.sym("obj_val", 1)
        self.alpha = 1.0
        self.alpha_reduction = 0.9

        penalty_term = ca.sum1(slack_variables)
        g = ca.vertcat(
            problem.g,
            # x_bin_var - rounded < slack
            # - (x_bin_var - rounded) < slack
            slack_variables - (x_bin_var - rounded_value),
            slack_variables + (x_bin_var - rounded_value),
        )
        options.update({"jit": s.WITH_JIT, "ipopt.max_iter": 5000})
        self.solver = ca.nlpsol(
            "nlpsol",
            "ipopt",
            {
                "f": alpha / obj_val * problem.f
                + (1 - alpha) / int_error * penalty_term,
                "g": g,
                "x": ca.vertcat(problem.x, slack_variables),
                "p": ca.vertcat(problem.p, rounded_value, alpha, int_error, obj_val),
            },
            options,
        )

    def solve(self, nlpdata: MinlpData, int_error, obj_val) -> MinlpData:
        success_out = []
        sols_out = []
        logger.info(f"Solving objective FP with alpha {self.alpha}")
        for sol in nlpdata.solutions_all:
            lbx = nlpdata.lbx
            ubx = nlpdata.ubx

            x_bin_var = to_0d(sol["x"][self.idx_x_integer])
            new_sol = self.solver(
                x0=ca.vertcat(nlpdata.x0, np.zeros(self.nr_x_bin)),
                lbx=ca.vertcat(lbx, np.zeros(self.nr_x_bin)),
                ubx=ca.vertcat(ubx, ca.inf * np.ones(self.nr_x_bin)),
                lbg=ca.vertcat(nlpdata.lbg, np.zeros(2 * self.nr_x_bin)),
                ubg=ca.vertcat(nlpdata.ubg, ca.inf * np.ones(2 * self.nr_x_bin)),
                p=ca.vertcat(
                    nlpdata.p, x_bin_var, np.array([self.alpha, int_error, obj_val])
                ),
            )

            success, _ = self.collect_stats("OFP", sol=new_sol)
            if not success:
                logger.debug("Infeasible solution!")
            success_out.append(success)
            sols_out.append(new_sol)

        self.alpha *= self.alpha_reduction
        nlpdata.prev_solutions = sols_out
        nlpdata.solved_all = success_out
        return nlpdata



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Test if all combinations work."""

import unittest
from os import path, makedirs, listdir
from shutil import rmtree
from camino.utils.cache import CachedFunction, ca


FILE_DIR = ".test"
FILE_PATH = path.join(FILE_DIR, "test")


def dummy_casadi_function():
    """Create a simple function."""
    x = ca.SX.sym("x", 2)
    f = x[0] ** 2 + x[1]
    return ca.Function("f", [x], [f])


class TestSolver(unittest.TestCase):
    """Test."""

    def setUp(self):
        """Setup a folder."""
        makedirs(FILE_DIR, exist_ok=True)

    def tearDown(self):
        """Remove cached files."""
        try:
            rmtree(FILE_DIR)
        except Exception:
            pass

    def test_cached_function(self):
        """Test a cached function."""
        f1 = CachedFunction("test", dummy_casadi_function, FILE_PATH, do_compile=False)
        out = f1([2.0, 3.0])
        self.assertEqual(out[0], 7.0)

        files = listdir(FILE_DIR)
        for f in files:
            self.assertFalse(f.endswith(".so"))

        f2 = CachedFunction("test", dummy_casadi_function, FILE_PATH, do_compile=False)
        out = f2([2.0, 3.0])
        self.assertEqual(out[0], 7.0)
        f2.f.jacobian().jacobian()

    def test_cached_function_compiled(self):
        """Test a cached function."""
        f1 = CachedFunction("test", dummy_casadi_function, FILE_PATH, do_compile=True)
        out = f1([2.0, 3.0])
        self.assertEqual(out[0], 7.0)

        files = listdir(FILE_DIR)
        has_so = False
        for f in files:
            if f.endswith(".so"):
                has_so = True

        self.assertTrue(has_so)

        f2 = CachedFunction("test", dummy_casadi_function, FILE_PATH, do_compile=True)
        out = f2([2.0, 3.0])
        self.assertEqual(out[0], 7.0)
        f2.f.jacobian().jacobian()


if __name__ == "__main__":
    unittest.main()



"""
Solar Thermal Climate System (STCS) at Karsruhe University of Applied Sciences.

Adrian Buerger, 2022
Adapted by Wim Van Roy and Andrea Ghezzi, 2023
"""

import numpy as np
from camino.settings import Settings
from camino.problems import MetaDataOcp
from camino.problems.solarsys.system import System, ca
from camino.problems.solarsys.ambient import Ambient, Timing
from camino.problems.solarsys.simulator import Simulator
from camino.problems.dsc import Description
from camino.settings import GlobalSettings
from camino.utils.cache import CachedFunction
import logging

from camino.utils.conversion import to_0d, convert_to_flat_list

logger = logging.getLogger(__name__)


def create_stcs_problem(simplified=False, with_slack=True):
    """Build problem."""
    logger.debug("Start processing")
    system = System()
    timing = Timing(simplified=simplified)
    ambient = Ambient(timing)
    dsc = Description()
    n_steps = timing.N

    # Run simulator and predictor and use those output to warm start
    simulator = Simulator(ambient=ambient, N=n_steps)
    simulator.solve()
    # x_hat = simulator.predict()

    collocation_nodes = 2

    logger.debug("Constructing bounds")
    x_min = system.p_op["T"]["min"] * np.ones((n_steps + 1, system.nx))
    x_max = system.p_op["T"]["max"] * np.ones((n_steps + 1, system.nx))
    x_max[:, system.x_index["T_shx_psc"][-1]] = system.p_op["T_sc"]["T_feed_max"]
    x_max[:, system.x_index["T_lts"]] = system.p_op["T_lts"]["max"]

    u_min = np.hstack(
        [
            system.p_op["v_ppsc"]["min_mpc"] * np.ones((n_steps, 1)),
            system.p_op["p_mpsc"]["min_mpc"] * np.ones((n_steps, 1)),
            system.p_op["v_pssc"]["min_mpc"] * np.ones((n_steps, 1)),
            -np.ones((n_steps, 1)),
            # The upcoming controls are constrained later in the NLP using inequality constraints
            np.zeros((n_steps, 1)),
            np.zeros((n_steps, 1)),
        ]
    )
    u_max = np.hstack(
        [
            system.p_op["v_ppsc"]["max"] * np.ones((n_steps, 1)),
            system.p_op["p_mpsc"]["max"] * np.ones((n_steps, 1)),
            system.p_op["v_pssc"]["max"] * np.ones((n_steps, 1)),
            np.ones((n_steps, 1)),
            np.inf * np.ones((n_steps, 1)),
            np.inf * np.ones((n_steps, 1)),
        ]
    )
    min_up_times = (
        np.asarray(system.p_op["acm"]["min_up_time"] + system.p_op["hp"]["min_up_time"])
        - 1e-3
    )
    min_down_times = (
        np.asarray(
            system.p_op["acm"]["min_down_time"] + system.p_op["hp"]["min_down_time"]
        )
        - 1e-3
    )

    logger.debug("Creating basic equations")
    F = system.get_system_dynamics_collocation(collocation_nodes)
    T_ac_min_fcn = CachedFunction("stcs", system.get_t_ac_min_function)
    T_ac_max_fcn = CachedFunction("stcs", system.get_t_ac_max_function)
    slacked_state_fcn = CachedFunction("stcs", system.get_slacked_state_fcn)
    v_ppsc_so_fpsc_fcn = CachedFunction("stcs", system.get_v_ppsc_so_fpsc_fcn)
    v_ppsc_so_vtsc_fcn = CachedFunction("stcs", system.get_v_ppsc_so_vtsc_fcn)
    v_ppsc_so_fcn = CachedFunction("stcs", system.get_v_ppsc_so_fcn)
    mdot_hts_b_max_fcn = CachedFunction("stcs", system.get_mdot_hts_b_max_fcn)
    electric_power_balance_fcn = CachedFunction(
        "stcs", system.get_electric_power_balance_fcn
    )
    F1_fcn = CachedFunction("stcs", system.get_F1_fcn)
    F2_fcn = system.get_F2_fcn()  # CachedFunction("stcs", system.get_F2_fcn)

    logger.debug("Create NLP problem")
    x_k_0 = dsc.add_parameters("x0", system.nx, to_0d(simulator.x_data[0, :]).tolist())
    u_k_prev = None
    tk = ambient.get_t0()
    F1 = []
    F2 = []

    for k in range(n_steps):
        logger.debug(f"Create NLP step {k}")
        dt = ambient.time_steps[k]
        tk += dt

        x_k_full = [x_k_0] + [
            dsc.sym(
                f"x_{k}_c",
                system.nx,
                lb=-np.inf,
                ub=np.inf,
                w0=to_0d(simulator.x_data[k, :]).tolist(),
            )
            for j in range(1, collocation_nodes + 1)
        ]
        x_k_next_0 = dsc.sym(
            "x",
            system.nx,
            lb=-ca.inf,
            ub=ca.inf,
            w0=to_0d(simulator.x_data[k + 1, :]).tolist(),
        )

        # Add new binary controls
        b_k = dsc.sym_bool("b", system.nb)

        # Add new continuous controls
        u_k = dsc.sym(
            "u",
            system.nu,
            lb=u_min[k, :],
            ub=u_max[k, :],
            w0=to_0d(simulator.u_data[k, :]).tolist(),
        )

        if u_k_prev is None:
            u_k_prev = u_k

        # Add new parametric controls
        params = convert_to_flat_list(
            system.nc, system.c_index, ambient.interpolate(tk - dt)
        )
        c_k = dsc.add_parameters("c", system.nc, params)
        dt_k = dsc.add_parameters("dt", 1, dt.total_seconds())

        # Add collocation equations
        F_k_inp = {"x_k_" + str(i): x_k_i for i, x_k_i in enumerate(x_k_full)}
        F_k_inp.update(
            {
                "x_k_next": x_k_next_0,
                "c_k": c_k,
                "u_k": u_k,
                "b_k": b_k,
                "dt_k": dt_k,
            }
        )
        F_k = F(**F_k_inp)
        dsc.eq(F_k["eq_c"], 0)
        dsc.eq(F_k["eq_d"], 0)

        # Add new slack variable for T_ac_min condition
        if with_slack:
            s_ac_lb_k = dsc.sym("s_ac_lb", system.n_s_ac_lb, 0, ca.inf)
        else:
            s_ac_lb_k = np.zeros((system.n_s_ac_lb,))

        # Setup T_ac_min conditions
        dsc.leq(0, T_ac_min_fcn(x_k_0, c_k, b_k, s_ac_lb_k))

        # Add new slack variable for T_ac_max condition
        if with_slack:
            s_ac_ub_k = dsc.sym("s_ac_ub", system.n_s_ac_ub, 0, ca.inf)
        else:
            s_ac_ub_k = np.zeros((system.n_s_ac_ub,))

        # Setup T_ac_max conditions
        dsc.leq(T_ac_max_fcn(x_k_0, c_k, b_k, s_ac_ub_k), 0)

        # Add new slack variable for state limits soft constraints
        if with_slack:
            s_x_k = dsc.sym("s_x", system.nx, -ca.inf, ca.inf, w0=0)
        else:
            s_x_k = np.zeros((system.nx,))

        # Setup state limits as soft constraints to prevent infeasibility
        dsc.add_g(
            x_min[k + 1, :], slacked_state_fcn(x_k_next_0, s_x_k), x_max[k + 1, :]
        )

        # Assure ppsc is running at high speed when collector temperature is high
        s_ppsc_k = dsc.sym("s_ppsc_fpsc", 1, lb=0, ub=1)

        dsc.leq(v_ppsc_so_fpsc_fcn(x_k_0, s_ppsc_k), 0)
        dsc.leq(v_ppsc_so_vtsc_fcn(x_k_0, s_ppsc_k), 0)
        dsc.leq(v_ppsc_so_fcn(u_k, s_ppsc_k), 0)

        # Assure HTS bottom layer mass flows are always smaller or equal to
        # the corresponding total pump flow
        dsc.leq(mdot_hts_b_max_fcn(u_k, b_k), 0)

        # Electric power balance
        dsc.eq(electric_power_balance_fcn(x_k_0, u_k, b_k, c_k), 0)

        # SOS1 constraint
        dsc.add_g(0, ca.sum1(b_k), 1)

        F1.append(
            np.sqrt(dt_k / 3600) * F1_fcn(s_ac_lb_k, s_ac_ub_k, s_x_k, u_k, u_k_prev)
        )
        F2.append((dt_k / 3600) * F2_fcn(u_k, c_k))

        x_k_0 = x_k_next_0
        u_k_prev = u_k

    # Concatenate objects
    F1 = 0.1 * ca.veccat(*F1)
    F2 = 0.01 * sum(F2)

    # Specify residual for GN Hessian computation
    dsc.r = F1

    idx_b_2d = np.asarray(dsc.get_indices("b")).T
    w = ca.vertcat(*dsc.w)

    # Set min up time constraints
    for k, dt in enumerate(ambient.time_steps):
        for i in range(system.nb):
            uptime = dt.total_seconds()
            it = 0
            # logger.debug(f"{k=}, {i=}, {dt=}")
            while uptime < min_up_times[i]:
                # logger.debug(f"{uptime=}")
                if k >= 0:
                    b_k = w[idx_b_2d[i, k]]
                else:
                    b_k = 0
                if (k - 1 >= 0) and (k - 1 < n_steps):
                    idx_k_1 = idx_b_2d[i, k - 1]
                    b_k_1 = w[idx_k_1]
                else:
                    b_k_1 = 0
                if (k - it - 2 >= 0) and (k - it - 2 < n_steps):
                    idx_k_dt = idx_b_2d[i, k - it - 2]
                    b_k_dt = w[idx_k_dt]
                else:
                    b_k_dt = 0
                dsc.leq(-b_k + b_k_1 - b_k_dt, 0, is_dwell_time=1)
                it += 1
                # logger.debug(f"{dsc.g[-1]}")
                uptime += ambient.time_steps[k - it].total_seconds()
    # Set min down time constraints
    for k, dt in enumerate(ambient.time_steps):
        for i in range(system.nb):
            downtime = dt.total_seconds()
            it = 0
            # logger.debug(f"{k=}, {i=}, {dt=}")
            while downtime < min_down_times[i]:
                # logger.debug(f"{downtime=}")
                if k >= 0:
                    b_k = w[idx_b_2d[i, k]]
                else:
                    b_k = 0
                if (k - 1 >= 0) and (k - 1 < n_steps):
                    idx_k_1 = idx_b_2d[i, k - 1]
                    b_k_1 = w[idx_k_1]
                else:
                    b_k_1 = 0
                if (k - it - 2 >= 0) and (k - it - 2 < n_steps):
                    idx_k_dt = idx_b_2d[i, k - it - 2]
                    b_k_dt = w[idx_k_dt]
                else:
                    b_k_dt = 0
                dsc.leq(b_k - b_k_1 + b_k_dt, 1, is_dwell_time=1)
                it += 1
                # logger.debug(f"{dsc.g[-1]}")
                downtime += ambient.time_steps[k - it].total_seconds()

    # Setup objective
    dsc.f = 0.5 * ca.mtimes(F1.T, F1) + F2
    logger.debug("NLP created")
    prob = dsc.get_problem()
    x_bar = GlobalSettings.CASADI_VAR.sym("x_bar", prob.x.shape)
    fun_F1 = ca.Function("F1", [prob.x, prob.p], [F1])
    fun_F2 = ca.Function("F2", [prob.x, prob.p], [F2])
    fun_grad_F1 = ca.Function("F1_grad", [prob.x, prob.p], [ca.jacobian(F1, prob.x).T])
    fun_grad_F2 = ca.Function("F2_grad", [prob.x, prob.p], [ca.jacobian(F2, prob.x).T])

    x = prob.x
    prob.f_qp = ca.Function(
        "F_qp",
        [x, x_bar, prob.p],
        [
            0.5 * ca.mtimes(fun_F1(x_bar, prob.p).T, fun_F1(x_bar, prob.p))
            + ca.mtimes(
                [(x - x_bar).T, fun_grad_F1(x_bar, prob.p), fun_F1(x_bar, prob.p)]
            )
            + 0.5
            * ca.mtimes(
                [
                    (x - x_bar).T,
                    fun_grad_F1(x_bar, prob.p),
                    fun_grad_F1(x_bar, prob.p).T,
                    (x - x_bar),
                ]
            )
            + fun_F2(x_bar, prob.p)
            + ca.mtimes(fun_grad_F2(x_bar, prob.p).T, x - x_bar)
        ],
    )
    meta = MetaDataOcp(
        n_state=system.nx,
        n_continuous_control=system.nu,
        n_discrete_control=system.nb,
        idx_param=dsc.indices_p,
        idx_state=np.hstack(dsc.indices["x"]).tolist(),
        idx_control=np.hstack(dsc.indices["u"]).tolist(),
        idx_bin_control=np.hstack(dsc.indices["b"]).tolist(),
        initial_state=to_0d(simulator.x_data[0, :]).tolist(),
        dt=ambient.time_steps,
        min_uptime=min_up_times,
        min_downtime=min_down_times,
    )
    prob.meta = meta
    data = dsc.get_data()
    data.x0[prob.idx_x_integer] = to_0d(simulator.b_data).flatten().tolist()
    s = Settings()
    s.BRMIQP_GAP = 0.1
    s.LBMILP_GAP = 0.05
    s.ALPHA_KRONQVIST = 0.5
    s.USE_RELAXED_SOL_AS_LINEARIZATION = True
    s.USE_TIGHT_MIPGAP_FIRST_ITERATION = True
    s.TIME_LIMIT = 0.5 * 3600
    s.IPOPT_SETTINGS.update(
        {
            "ipopt.linear_solver": "ma57",
            "ipopt.max_cpu_time": 300,
            "ipopt.mumps_mem_percent": 10000,
            "ipopt.mumps_pivtol": 0.001,
            "ipopt.max_cpu_time": 3600.0,
            "ipopt.max_iter": 5000,
            "ipopt.acceptable_tol": 1e-1,
            "ipopt.acceptable_iter": 8,
            "ipopt.acceptable_constr_viol_tol": 10.0,
            "ipopt.acceptable_dual_inf_tol": 10.0,
            "ipopt.acceptable_compl_inf_tol": 10.0,
            "ipopt.acceptable_obj_change_tol": 1e-1,
            "ipopt.mu_strategy": "adaptive",
            "ipopt.mu_target": 1e-5,
            "ipopt.print_frequency_iter": 100,
        }
    )
    s.MIP_SETTINGS_ALL["gurobi"].update(
        {
            "gurobi.MIPGap": 0.1,
            "gurobi.FeasibilityTol": s.CONSTRAINT_INT_TOL,
            "gurobi.IntFeasTol": s.CONSTRAINT_INT_TOL,
            "gurobi.Heuristics": 0.05,
            "gurobi.PoolSearchMode": 0,
            "gurobi.PoolSolutions": 3,
            "gurobi.TimeLimit": 600,
            "gurobi.NodeMethod": 2,
            "gurobi.Method": 2,
        }
    )
    s.BONMIN_SETTINGS = {
        "bonmin.time_limit": s.TIME_LIMIT,
        "bonmin.tree_search_strategy": "dive",
        "bonmin.node_comparison": "best-bound",
        "bonmin.allowable_fraction_gap": Settings.MINLP_TOLERANCE,
        "bonmin.allowable_gap": Settings.MINLP_TOLERANCE_ABS,
        "bonmin.constr_viol_tol": s.CONSTRAINT_TOL,
        "bonmin.linear_solver": "ma57",
    }

    # Improve calculation speed by getting indices:
    # set_constraint_types(prob, *cache_data(
    #     f"scts_{n_steps}_{with_slack}", inspect_problem, prob, data
    # ))

    return prob, data, s


if __name__ == "__main__":
    from camino.utils import setup_logger, logging
    from camino.utils.validate import check_solution
    from camino.stats import Stats
    from camino.solvers.subsolvers.nlp import NlpSolver
    from datetime import datetime
    import pickle

    setup_logger(logging.DEBUG)
    prob, data, s = create_stcs_problem()
    stats = Stats(
        mode="custom",
        problem_name="stcs",
        datetime=datetime.now().strftime("%Y-%m-%d_%H:%M:%S"),
        data={},
    )
    nlp = NlpSolver(prob, stats, s)

    # with open("data/nlpargs_adrian.pickle", 'rb') as f:
    #     nlpargs_adrian = pickle.load(f)
    # data.x0 = nlpargs_adrian['x0']
    # data.p = nlpargs_adrian['p']
    # data.lbx = nlpargs_adrian['lbx']
    # data.ubx = nlpargs_adrian['ubx']
    # data.lbg = nlpargs_adrian['lbg']
    # data.ubg = nlpargs_adrian['ubg']

    data = nlp.solve(data, integers_relaxed=True)  # solve relaxed problem
    with open("results/x_star_rel_test.pickle", "wb") as f:
        pickle.dump(data.x_sol, f)
    breakpoint()
    check_solution(prob, data, data.prev_solution["x"], s)



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""A MIP solver."""

import casadi as ca
import logging
from camino.solvers import (
    MiSolverClass,
    Stats,
    MinlpProblem,
    MinlpData,
    regularize_options,
)
from camino.settings import Settings

logger = logging.getLogger(__name__)


class MipSolver(MiSolverClass):
    """
    Create MIP solver.

    This solver solves a mixed-integer problem (MIP).
    Constraints must be linear, for linear objective it solves a MILP, for quadratic
    cost it solves a MIQP.
    """

    def __init__(
        self, problem: MinlpProblem, data: MinlpData, stats: Stats, s: Settings
    ):
        """Create NLP problem."""
        super(MipSolver, self).__init__(problem, data, stats, s)
        self.options = regularize_options(s.MIP_SETTINGS, {}, s)

        self.idx_x_integer = problem.idx_x_integer
        self.nr_x = problem.x.shape[0]

        self.options["discrete"] = [
            1 if i in self.idx_x_integer else 0 for i in range(self.nr_x)
        ]

        self.solver = ca.qpsol(
            "mip_solver",
            self.settings.MIP_SOLVER,
            {
                "f": problem.f,
                "g": problem.g,
                "x": problem.x,
                "p": problem.p,
            },
            self.options,
        )

    def solve(self, nlpdata: MinlpData) -> MinlpData:
        """Solve given MIP."""

        nlpdata.prev_solution = self.solver(
            x0=nlpdata.x0,
            lbx=nlpdata.lbx,
            ubx=nlpdata.ubx,
            lbg=nlpdata.lbg,
            ubg=nlpdata.ubg,
            p=nlpdata.p,
        )

        nlpdata.solved, stats = self.collect_stats(
            "MIP", sol=nlpdata.prev_solutions[-1]
        )
        return nlpdata

    def reset(self, nlpdata: MinlpData):
        """Reset problem data."""
        logger.warning("reset method not available for MipSolver class!")

    def warmstart(self, nlpdata: MinlpData):
        """Warmstart the algorithm."""
        logger.warning(
            "warmstart method not available for MipSolver class. To pass a linearization point use 'nlpdata.x0'."
        )



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""Utilities for generating plots."""

import numpy as np
from camino.data import MinlpData
from camino.problem import MinlpProblem
from camino.utils.conversion import to_0d


def get_control_vector(problem: MinlpProblem, data: MinlpData):
    out = []
    if problem.meta.n_continuous_control > 0:
        out.append(
            to_0d(data.x_sol)[problem.meta.idx_control].reshape(
                -1, problem.meta.n_continuous_control
            )
        )
    if problem.meta.n_discrete_control > 0:
        out.append(
            to_0d(data.x_sol)[problem.meta.idx_bin_control].reshape(
                -1, problem.meta.n_discrete_control
            )
        )
    if len(out) > 1:
        out = np.hstack(out)
    return out



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

import numpy as np
import casadi as ca
from camino.settings import Settings
from camino.data import MinlpData
from camino.problem import MinlpProblem
from camino.stats import Stats
from camino.utils import toc, logging
from camino.utils.conversion import to_0d


def almost_equal(a, b, EPS=1e-5):
    """Check if almost equal."""
    return a + EPS > b and a - EPS < b


class Constraints:
    """Store bounds."""

    def __init__(self, nr=0, eq=None, lb=None, ub=None):
        """Store bounds."""
        if nr == 0:
            eq, lb, ub = [], [], []

        self.nr = nr
        self.eq = eq
        self.lb = lb
        self.ub = ub
        self.a = None
        self.b = None

    def add(self, lb, eq, ub):
        """Add a bound."""
        if self.nr == 0:
            self.nr += 1
            self.eq = eq
            self.lb = ca.DM(lb)
            self.ub = ca.DM(ub)
        else:
            self.nr += 1
            self.eq = ca.vertcat(self.eq, eq)
            self.lb = ca.vertcat(self.lb, lb)
            self.ub = ca.vertcat(self.ub, ub)

    def to_generic(self):
        """Convert to a generic class."""
        return self

    def __add__(self, other):
        """Add two bounds."""
        other = other.to_generic()
        if other.nr == 0:
            return self
        if self.nr == 0:
            return other

        return Constraints(
            self.nr + other.nr,
            ca.vertcat(self.eq, other.eq),
            ca.vertcat(self.lb, other.lb),
            ca.vertcat(self.ub, other.ub),
        )

    def get_a(self, x, nr_x):
        """Get A matrix."""
        if self.a is None:
            self.a = ca.Function("geq", [x], [ca.jacobian(self.eq, x)])(
                np.zeros((nr_x,))
            )
        return self.a

    def get_b(self, x, nr_x):
        """Get B matrix."""
        if self.b is None:
            self.b = ca.Function("eq", [x], [self.eq])(np.zeros((nr_x,)))
        return self.b

    def __str__(self):
        """Represent."""
        out = f"Eq: {self.nr}\n\n"
        for i in range(self.nr):
            out += f"{self.lb[i]} <= {self.eq[i]} <= {self.ub[i]}\n"
        return out


def bin_equal(sol1, sol2, idx_x_integer):
    """Binary variables equal."""
    return np.allclose(
        sol1[idx_x_integer], sol2[idx_x_integer], equal_nan=False, atol=1e-2
    )


def any_equal(sol, refs, idx_x_integer):
    """Check if any is equal."""
    for ref in refs:
        if bin_equal(sol, ref, idx_x_integer):
            return True
    return False


def get_solutions_pool(nlpdata, success, stats, s: Settings, solution, idx_x_integer):
    """Get pool of solutions if exists."""
    if s.USE_SOLUTION_POOL and stats and "pool_sol_nr" in stats:
        sols = [solution]
        x_sols = [solution["x"]]

        for i in range(1, stats["pool_sol_nr"]):
            x = ca.DM(stats["pool_solutions"][i])
            if not any_equal(x, x_sols, idx_x_integer):
                sols.append({"f": stats["pool_obj_val"][i], "x": x})
                x_sols.append(x)
        nlpdata.prev_solutions = sols
        nlpdata.solved_all = [success for i in sols]
    else:
        nlpdata.prev_solutions = [solution]
        nlpdata.solved_all = [success]

    return nlpdata


def get_termination_condition(
    termination_type, problem: MinlpProblem, data: MinlpData, s: Settings
):
    """
    Get termination condition.

    :param termination_type: String of the termination type (gradient, std or equality)
    :param problem: problem
    :param data: data
    :return: callable that returns true if the termination condition holds
    """

    def max_time(ret, s, stats):
        done = False
        if s.TIME_LIMIT_SOLVER_ONLY:
            done = stats["t_solver_total"] > s.TIME_LIMIT or toc() > s.TIME_LIMIT * 3
        else:
            done = toc() > s.TIME_LIMIT

        if done:
            logging.info("Terminated - TIME LIMIT")
            return True
        return ret

    if termination_type == "gradient":
        idx_x_integer = problem.idx_x_integer
        f_fn = ca.Function(
            "f", [problem.x, problem.p], [problem.f], {"jit": s.WITH_JIT}
        )
        grad_f_fn = ca.Function(
            "gradient_f_x",
            [problem.x, problem.p],
            [ca.gradient(problem.f, problem.x)],
            {"jit": s.WITH_JIT},
        )

        def func(
            stats: Stats, s: Settings, lb=None, ub=None, x_best=None, x_current=None
        ):
            ret = (
                to_0d(
                    f_fn(x_current, data.p)
                    + grad_f_fn(x_current, data.p)[idx_x_integer].T
                    @ (x_current[idx_x_integer] - x_best[idx_x_integer])
                    - f_fn(x_best, data.p)
                )
                >= 0
            )
            if ret:
                logging.info("Terminated - gradient ok")
            return max_time(ret, s, stats)

    elif termination_type == "equality":
        idx_x_integer = problem.idx_x_integer

        def func(
            stats: Stats, s: Settings, lb=None, ub=None, x_best=None, x_current=None
        ):
            if isinstance(x_best, list):
                for x in x_best:
                    if np.allclose(
                        x[idx_x_integer],
                        x_current[idx_x_integer],
                        equal_nan=False,
                        atol=s.EPS,
                    ):
                        logging.info(f"Terminated - all close within {s.EPS}")
                        return True
                return max_time(False, s, stats)
            else:
                ret = np.allclose(
                    x_best[idx_x_integer],
                    x_current[idx_x_integer],
                    equal_nan=False,
                    atol=s.EPS,
                )
                if ret:
                    logging.info(f"Terminated - all close within {s.EPS}")
                return max_time(ret, s, stats)

    elif termination_type == "std":

        def func(
            stats: Stats, s: Settings, lb=None, ub=None, x_best=None, x_current=None
        ):
            tol_abs = max(
                (abs(lb) + abs(ub)) * s.MINLP_TOLERANCE / 2, s.MINLP_TOLERANCE_ABS
            )
            ret = (lb + tol_abs - ub) >= 0
            if ret:
                logging.info(f"Terminated: {lb} >= {ub} - {tol_abs} ({tol_abs})")
            else:
                logging.info(f"Not Terminated: {lb} <= {ub} - {tol_abs} ({tol_abs})")
            return max_time(ret, s, stats)

    else:
        raise AttributeError(
            f"Invalid type of termination condition, you set '{termination_type}' but the only option is 'std'!"
        )
    return func



# This file is part of CAMINO
# Copyright (C) 2024  Andrea Ghezzi, Wim Van Roy, Sebastian Sager, Moritz Diehl
# SPDX-License-Identifier: GPL-3.0-or-later

"""
Collection of Pumps.

This folder contains a collection of solvers based on the 'feasibility pump' idea.
They include:
    - Feasibility Pump
    - Objective Feasibility Pump
    - Random Objective Feasibility Pump
"""

from camino.settings import Settings
from camino.stats import Stats
from camino.data import MinlpData
from camino.problem import MinlpProblem
from camino.solvers.pumps.projections import (
    LinearProjection,
    ObjectiveLinearProjection,
)
from camino.solvers.pumps.projection_random import RandomDirectionProjection
from camino.solvers.pumps.base import PumpBase
from camino.solvers.pumps.base_random import PumpBaseRandom


class FeasibilityPump(PumpBase):
    """
    Feasibility Pump

    According to:
        Bertacco, L., Fischetti, M., & Lodi, A. (2007). A feasibility pump heuristic for general mixed-integer problems.
        Discrete Optimization, 4(1), 63-76.
    """

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        nlp=None,
    ):
        """Create a solver class."""
        pump = LinearProjection(problem, stats, settings)
        super(FeasibilityPump, self).__init__(
            problem, data, stats, settings, pump, nlp=nlp
        )


class ObjectiveFeasibilityPump(PumpBase):
    """
    Objective Feasibility Pump

        Sharma, S., Knudsen, B. R., & Grimstad, B. (2016). Towards an objective feasibility pump for convex MINLPs.
        Computational Optimization and Applications, 63, 737-753.
    """

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        nlp=None,
    ):
        """Create a solver class."""
        pump = ObjectiveLinearProjection(problem, stats, settings)
        super(ObjectiveFeasibilityPump, self).__init__(
            problem, data, stats, settings, pump, nlp=nlp
        )


class RandomObjectiveFeasibilityPump(PumpBaseRandom):
    """
    Random objective pump

    Alternative version of the objective pump where the weights are randomly chosen. This allows to remove some other
    random recovery mechanisms.
    """

    def __init__(
        self,
        problem: MinlpProblem,
        data: MinlpData,
        stats: Stats,
        settings: Settings,
        nlp=None,
        norm=2,
        penalty_scaling=0.5,
    ):
        """Create a solver class."""
        pump = RandomDirectionProjection(
            problem, stats, settings, norm=norm, penalty_scaling=penalty_scaling
        )
        super(RandomObjectiveFeasibilityPump, self).__init__(
            problem, data, stats, settings, pump, nlp=nlp
        )
